from collections import defaultdict, OrderedDict
from typing import Generator, Dict, Any, Optional, Set, List
import operator
import logging

import networkx

import claripy
import ailment

from angr.utils.graph import GraphUtils
from ...utils.lazy_import import lazy_import
from ...utils import is_pyinstaller
from ...utils.graph import dominates, inverted_idoms
from ...block import Block, BlockNode
from .peephole_optimizations import InvertNegatedLogicalConjunctionsAndDisjunctions
from .structuring.structurer_nodes import (
    MultiNode,
    EmptyBlockNotice,
    SequenceNode,
    CodeNode,
    SwitchCaseNode,
    BreakNode,
    ConditionalBreakNode,
    LoopNode,
    ConditionNode,
    ContinueNode,
    CascadingConditionNode,
    IncompleteSwitchCaseNode,
)
from .graph_region import GraphRegion
from .utils import first_nonlabel_statement, peephole_optimize_expr

if is_pyinstaller():
    # PyInstaller is not happy with lazy import
    import sympy
else:
    sympy = lazy_import("sympy")


l = logging.getLogger(__name__)


_UNIFIABLE_COMPARISONS = {
    "__ne__",
    "__gt__",
    "__ge__",
    "UGT",
    "UGE",
    "SGT",
    "SGE",
}

#
# Util methods and mapping used during AIL AST to claripy AST conversion
#


def _op_with_unified_size(op, conv, operand0, operand1):
    # ensure operand1 is of the same size as operand0
    if isinstance(operand1, ailment.Expr.Const):
        # amazing - we do the eazy thing here
        return op(conv(operand0, nobool=True), operand1.value)
    if operand1.bits == operand0.bits:
        return op(conv(operand0, nobool=True), conv(operand1))
    # extension is required
    assert operand1.bits < operand0.bits
    operand1 = ailment.Expr.Convert(None, operand1.bits, operand0.bits, False, operand1)
    return op(conv(operand0, nobool=True), conv(operand1, nobool=True))


def _dummy_bvs(condition, condition_mapping, name_suffix=""):
    var = claripy.BVS(f"ailexpr_{repr(condition)}{name_suffix}", condition.bits, explicit_name=True)
    condition_mapping[var.args[0]] = condition
    return var


_ail2claripy_op_mapping = {
    "LogicalAnd": lambda expr, conv, _: claripy.And(conv(expr.operands[0]), conv(expr.operands[1])),
    "LogicalOr": lambda expr, conv, _: claripy.Or(conv(expr.operands[0]), conv(expr.operands[1])),
    "CmpEQ": lambda expr, conv, _: conv(expr.operands[0]) == conv(expr.operands[1]),
    "CmpNE": lambda expr, conv, _: conv(expr.operands[0]) != conv(expr.operands[1]),
    "CmpLE": lambda expr, conv, _: conv(expr.operands[0]) <= conv(expr.operands[1]),
    "CmpLEs": lambda expr, conv, _: claripy.SLE(conv(expr.operands[0]), conv(expr.operands[1])),
    "CmpLT": lambda expr, conv, _: conv(expr.operands[0]) < conv(expr.operands[1]),
    "CmpLTs": lambda expr, conv, _: claripy.SLT(conv(expr.operands[0]), conv(expr.operands[1])),
    "CmpGE": lambda expr, conv, _: conv(expr.operands[0]) >= conv(expr.operands[1]),
    "CmpGEs": lambda expr, conv, _: claripy.SGE(conv(expr.operands[0]), conv(expr.operands[1])),
    "CmpGT": lambda expr, conv, _: conv(expr.operands[0]) > conv(expr.operands[1]),
    "CmpGTs": lambda expr, conv, _: claripy.SGT(conv(expr.operands[0]), conv(expr.operands[1])),
    "Add": lambda expr, conv, _: conv(expr.operands[0], nobool=True) + conv(expr.operands[1], nobool=True),
    "Sub": lambda expr, conv, _: conv(expr.operands[0], nobool=True) - conv(expr.operands[1], nobool=True),
    "Mul": lambda expr, conv, _: conv(expr.operands[0], nobool=True) * conv(expr.operands[1], nobool=True),
    "Div": lambda expr, conv, _: conv(expr.operands[0], nobool=True) / conv(expr.operands[1], nobool=True),
    "Mod": lambda expr, conv, _: conv(expr.operands[0], nobool=True) % conv(expr.operands[1], nobool=True),
    "Not": lambda expr, conv, _: claripy.Not(conv(expr.operand)),
    "Neg": lambda expr, conv, _: -conv(expr.operand),
    "BitwiseNeg": lambda expr, conv, _: ~conv(expr.operand),
    "Xor": lambda expr, conv, _: conv(expr.operands[0], nobool=True) ^ conv(expr.operands[1], nobool=True),
    "And": lambda expr, conv, _: conv(expr.operands[0], nobool=True) & conv(expr.operands[1], nobool=True),
    "Or": lambda expr, conv, _: conv(expr.operands[0], nobool=True) | conv(expr.operands[1], nobool=True),
    "Shr": lambda expr, conv, _: _op_with_unified_size(claripy.LShR, conv, expr.operands[0], expr.operands[1]),
    "Shl": lambda expr, conv, _: _op_with_unified_size(operator.lshift, conv, expr.operands[0], expr.operands[1]),
    "Sar": lambda expr, conv, _: _op_with_unified_size(operator.rshift, conv, expr.operands[0], expr.operands[1]),
    # There are no corresponding claripy operations for the following operations
    "DivMod": lambda expr, _, m: _dummy_bvs(expr, m),
    "CmpF": lambda expr, _, m: _dummy_bvs(expr, m),
    "Mull": lambda expr, _, m: _dummy_bvs(expr, m),
    "Mulls": lambda expr, _, m: _dummy_bvs(expr, m),
    "Reinterpret": lambda expr, _, m: _dummy_bvs(expr, m),
}

#
# The ConditionProcessor class
#


class ConditionProcessor:
    """
    Convert between claripy AST and AIL expressions. Also calculates reaching conditions of all nodes on a graph.
    """

    def __init__(self, arch, condition_mapping=None):
        self.arch = arch
        self._condition_mapping: Dict[str, Any] = {} if condition_mapping is None else condition_mapping
        self.jump_table_conds: Dict[int, Set] = defaultdict(set)
        self.edge_conditions = {}
        self.reaching_conditions = {}
        self.guarding_conditions = {}
        self._ast2annotations = {}

        self._peephole_expr_optimizations = [
            cls(None, None, None) for cls in [InvertNegatedLogicalConjunctionsAndDisjunctions]
        ]

    def clear(self):
        self._condition_mapping = {}
        self.jump_table_conds = defaultdict(set)
        self.reaching_conditions = {}
        self.guarding_conditions = {}
        self._ast2annotations = {}

    def recover_edge_condition(self, graph: networkx.DiGraph, src, dst):
        edge = src, dst
        edge_data = graph.get_edge_data(*edge)
        edge_type = edge_data.get("type", "transition") if edge_data is not None else "transition"
        try:
            predicate = self._extract_predicate(src, dst, edge_type)
        except EmptyBlockNotice:
            # catch empty block notice - although this should not really happen
            predicate = claripy.true
        return predicate

    def recover_edge_conditions(self, region, graph=None) -> Dict:
        edge_conditions = {}
        # traverse the graph to recover the condition for each edge
        graph = graph or region.graph
        for src in graph.nodes():
            nodes = list(graph[src])
            if len(nodes) >= 1:
                for dst in nodes:
                    predicate = self.recover_edge_condition(graph, src, dst)
                    edge_conditions[(src, dst)] = predicate

        self.edge_conditions = edge_conditions

    def recover_reaching_conditions(
        self, region, graph=None, with_successors=False, case_entry_to_switch_head: Optional[Dict[int, int]] = None
    ):
        def _strictly_postdominates(inv_idoms, node_a, node_b):
            """
            Does node A strictly post-dominate node B on the graph?
            """
            return dominates(inv_idoms, node_a, node_b)

        self.recover_edge_conditions(region, graph=graph)
        edge_conditions = self.edge_conditions

        if graph:
            _g = graph
            head = [node for node in graph.nodes if graph.in_degree(node) == 0][0]
        else:
            if with_successors and region.graph_with_successors is not None:
                _g = region.graph_with_successors
            else:
                _g = region.graph
            head = region.head

        # special handling for jump table entries - do not allow crossing between cases
        if case_entry_to_switch_head:
            _g = self._remove_crossing_edges_between_cases(_g, case_entry_to_switch_head)

        inverted_graph, idoms = inverted_idoms(_g)

        reaching_conditions = {}
        # recover the reaching condition for each node
        sorted_nodes = GraphUtils.quasi_topological_sort_nodes(_g)
        terminating_nodes = []
        for node in sorted_nodes:
            # create special conditions for all nodes that are jump table entries
            if case_entry_to_switch_head:
                if node.addr in case_entry_to_switch_head:
                    jump_target_var = self.create_jump_target_var(case_entry_to_switch_head[node.addr])
                    cond = jump_target_var == claripy.BVV(node.addr, self.arch.bits)
                    reaching_conditions[node] = cond
                    self.jump_table_conds[case_entry_to_switch_head[node.addr]].add(cond)
                    continue

            preds = _g.predecessors(node)
            reaching_condition = None

            out_degree = _g.out_degree(node)
            if out_degree == 0:
                terminating_nodes.append(node)

            if node is head:
                # the head is always reachable
                reaching_condition = claripy.true
            elif idoms is not None and _strictly_postdominates(idoms, node, head):
                # the node that post dominates the head is always reachable
                reaching_conditions[node] = claripy.true
            else:
                for pred in preds:
                    edge = (pred, node)
                    pred_condition = reaching_conditions.get(pred, claripy.true)
                    edge_condition = edge_conditions.get(edge, claripy.true)

                    if reaching_condition is None:
                        reaching_condition = claripy.And(pred_condition, edge_condition)
                    else:
                        reaching_condition = claripy.Or(claripy.And(pred_condition, edge_condition), reaching_condition)

            if reaching_condition is not None:
                reaching_conditions[node] = self.simplify_condition(reaching_condition)

        # My hypothesis: for nodes where two paths come together *and* those that cannot be further structured into
        # another if-else construct (we take the short-cut by testing if the operator is an "Or" after running our
        # condition simplifiers previously), we are better off using their "guarding conditions" instead of their
        # reaching conditions for if-else. see my super long chatlog with rhelmot on 5/14/2021.
        guarding_conditions = {}
        for the_node in sorted_nodes:
            preds = list(_g.predecessors(the_node))
            if len(preds) != 2:
                continue
            # generate a graph slice that goes from the region head to this node
            slice_nodes = list(networkx.dfs_tree(inverted_graph, the_node))
            subgraph = networkx.subgraph(_g, slice_nodes)
            # figure out which paths cause the divergence from this node
            nodes_do_not_reach_the_node = set()
            for node_ in subgraph:
                if node_ is the_node:
                    continue
                for succ in _g.successors(node_):
                    if not networkx.has_path(_g, succ, the_node):
                        nodes_do_not_reach_the_node.add(succ)

            diverging_conditions = []

            for node_ in nodes_do_not_reach_the_node:
                preds_ = list(_g.predecessors(node_))
                for pred_ in preds_:
                    if pred_ in nodes_do_not_reach_the_node:
                        continue
                    # this predecessor is the diverging node!
                    edge_ = pred_, node_
                    edge_condition = edge_conditions.get(edge_, None)
                    if edge_condition is not None:
                        diverging_conditions.append(edge_condition)

            if diverging_conditions:
                # the negation of the union of diverging conditions is the guarding condition for this node
                cond = claripy.Or(*map(claripy.Not, diverging_conditions))  # pylint:disable=bad-builtin
                guarding_conditions[the_node] = cond

        self.reaching_conditions = reaching_conditions
        self.guarding_conditions = guarding_conditions

    def remove_claripy_bool_asts(self, node, memo=None):
        # Convert claripy Bool ASTs to AIL expressions

        if memo is None:
            memo = {}

        if isinstance(node, SequenceNode):
            new_nodes = []
            for n in node.nodes:
                new_node = self.remove_claripy_bool_asts(n, memo=memo)
                new_nodes.append(new_node)
            new_seq_node = SequenceNode(node.addr, new_nodes)
            return new_seq_node

        elif isinstance(node, MultiNode):
            new_nodes = []
            for n in node.nodes:
                new_node = self.remove_claripy_bool_asts(n, memo=memo)
                new_nodes.append(new_node)
            new_multinode = MultiNode(nodes=new_nodes)
            return new_multinode

        elif isinstance(node, CodeNode):
            node = CodeNode(
                self.remove_claripy_bool_asts(node.node, memo=memo),
                None
                if node.reaching_condition is None
                else self.convert_claripy_bool_ast(node.reaching_condition, memo=memo),
            )
            return node

        elif isinstance(node, ConditionalBreakNode):
            return ConditionalBreakNode(
                node.addr,
                self.convert_claripy_bool_ast(node.condition, memo=memo),
                node.target,
            )

        elif isinstance(node, ConditionNode):
            return ConditionNode(
                node.addr,
                None
                if node.reaching_condition is None
                else self.convert_claripy_bool_ast(node.reaching_condition, memo=memo),
                self.convert_claripy_bool_ast(node.condition, memo=memo),
                self.remove_claripy_bool_asts(node.true_node, memo=memo),
                self.remove_claripy_bool_asts(node.false_node, memo=memo),
            )

        elif isinstance(node, CascadingConditionNode):
            cond_and_nodes = []
            for cond, child_node in node.condition_and_nodes:
                cond_and_nodes.append(
                    (
                        self.convert_claripy_bool_ast(cond, memo=memo),
                        self.remove_claripy_bool_asts(child_node, memo=memo),
                    )
                )
            else_node = None if node.else_node is None else self.remove_claripy_bool_asts(node.else_node, memo=memo)
            return CascadingConditionNode(
                node.addr,
                cond_and_nodes,
                else_node=else_node,
            )

        elif isinstance(node, LoopNode):
            result = node.copy()
            result.condition = (
                self.convert_claripy_bool_ast(node.condition, memo=memo) if node.condition is not None else None
            )
            result.sequence_node = self.remove_claripy_bool_asts(node.sequence_node, memo=memo)
            return result

        elif isinstance(node, SwitchCaseNode):
            return SwitchCaseNode(
                self.convert_claripy_bool_ast(node.switch_expr, memo=memo),
                OrderedDict(
                    (idx, self.remove_claripy_bool_asts(case_node, memo=memo)) for idx, case_node in node.cases.items()
                ),
                self.remove_claripy_bool_asts(node.default_node, memo=memo),
                addr=node.addr,
            )

        elif isinstance(node, IncompleteSwitchCaseNode):
            return IncompleteSwitchCaseNode(
                node.addr,
                self.remove_claripy_bool_asts(node.head, memo=memo),
                [self.remove_claripy_bool_asts(case, memo=memo) for case in node.cases],
            )

        else:
            return node

    @classmethod
    def get_last_statement(cls, block):
        """
        This is the buggy version of get_last_statements, because, you know, there can always be more than one last
        statement due to the existence of branching statements (like, If-then-else). All methods using
        get_last_statement() should switch to get_last_statements() and properly handle multiple last statements.
        """
        if type(block) is SequenceNode:
            if block.nodes:
                return cls.get_last_statement(block.nodes[-1])
            raise EmptyBlockNotice()
        if type(block) is CodeNode:
            return cls.get_last_statement(block.node)
        if type(block) is ailment.Block:
            if not block.statements:
                raise EmptyBlockNotice()
            return block.statements[-1]
        if type(block) is Block:
            raise NotImplementedError()
        if type(block) is BlockNode:
            raise NotImplementedError()
        if type(block) is MultiNode:
            # get the last node
            for the_block in reversed(block.nodes):
                try:
                    last_stmt = cls.get_last_statement(the_block)
                    return last_stmt
                except EmptyBlockNotice:
                    continue
            raise EmptyBlockNotice()
        if type(block) is LoopNode:
            return cls.get_last_statement(block.sequence_node)
        if type(block) is ConditionalBreakNode:
            return None
        if type(block) is ConditionNode:
            s = None
            if block.true_node:
                try:
                    s = cls.get_last_statement(block.true_node)
                except EmptyBlockNotice:
                    s = None
            if s is None and block.false_node:
                s = cls.get_last_statement(block.false_node)
            return s
        if type(block) is CascadingConditionNode:
            s = None
            if block.else_node is not None:
                s = cls.get_last_statement(block.else_node)
            else:
                for _, node in reversed(block.condition_and_nodes):
                    s = cls.get_last_statement(node)
                    if s is not None:
                        break
            return s
        if type(block) is BreakNode:
            return None
        if type(block) is ContinueNode:
            return None
        if type(block) is SwitchCaseNode:
            return None
        if type(block) is IncompleteSwitchCaseNode:
            return None
        if type(block) is GraphRegion:
            # normally this should not happen. however, we have test cases that trigger this case.
            return None

        raise NotImplementedError()

    @classmethod
    def get_last_statements(cls, block) -> List[Optional[ailment.Stmt.Statement]]:
        if type(block) is SequenceNode:
            for last_node in reversed(block.nodes):
                try:
                    last_stmts = cls.get_last_statements(last_node)
                    return last_stmts
                except EmptyBlockNotice:
                    # the node is empty. try the next one
                    continue

            raise EmptyBlockNotice()

        if type(block) is CodeNode:
            return cls.get_last_statements(block.node)
        if type(block) is ailment.Block:
            if not block.statements:
                raise EmptyBlockNotice()
            return [block.statements[-1]]
        if type(block) is Block:
            raise NotImplementedError()
        if type(block) is BlockNode:
            raise NotImplementedError()
        if type(block) is MultiNode:
            # get the last node
            for the_block in reversed(block.nodes):
                try:
                    last_stmts = cls.get_last_statements(the_block)
                    return last_stmts
                except EmptyBlockNotice:
                    continue
            raise EmptyBlockNotice()
        if type(block) is LoopNode:
            return cls.get_last_statements(block.sequence_node)
        if type(block) is ConditionalBreakNode:
            return [block]
        if type(block) is ConditionNode:
            s = []
            if block.true_node:
                try:
                    last_stmts = cls.get_last_statements(block.true_node)
                    s.extend(last_stmts)
                except EmptyBlockNotice:
                    pass
            else:
                s.append(None)
            if block.false_node:
                last_stmts = cls.get_last_statements(block.false_node)
                s.extend(last_stmts)
            else:
                s.append(None)
            return s
        if type(block) is CascadingConditionNode:
            s = []
            if block.else_node is not None:
                try:
                    last_stmts = cls.get_last_statements(block.else_node)
                    s.extend(last_stmts)
                except EmptyBlockNotice:
                    pass
            else:
                s.append(None)
            for _, node in block.condition_and_nodes:
                last_stmts = cls.get_last_statements(node)
                s.extend(last_stmts)
            return s
        if type(block) is BreakNode:
            return [block]
        if type(block) is ContinueNode:
            return [block]
        if type(block) is SwitchCaseNode:
            s = []
            for case in block.cases.values():
                s.extend(cls.get_last_statements(case))
            if block.default_node is not None:
                s.extend(cls.get_last_statements(block.default_node))
            else:
                s.append(None)
            return s
        if type(block) is IncompleteSwitchCaseNode:
            s = []
            for case in block.cases:
                s.extend(cls.get_last_statements(case))
            return s
        if type(block) is GraphRegion:
            # normally this should not happen. however, we have test cases that trigger this case.
            return []

        raise NotImplementedError()

    #
    # Path predicate
    #

    EXC_COUNTER = 1000

    def _extract_predicate(self, src_block, dst_block, edge_type) -> claripy.ast.Bool:
        if edge_type == "exception":
            # TODO: THIS IS ABSOLUTELY A HACK. AT THIS MOMENT YOU SHOULD NOT ATTEMPT TO MAKE SENSE OF EXCEPTION EDGES.
            self.EXC_COUNTER += 1
            return self.claripy_ast_from_ail_condition(
                ailment.Expr.BinaryOp(
                    None,
                    "CmpEQ",
                    (
                        ailment.Expr.Register(0x400000 + self.EXC_COUNTER, None, self.EXC_COUNTER, 64),
                        ailment.Expr.Const(None, None, self.EXC_COUNTER, 64),
                    ),
                    False,
                ),
            )

        if type(src_block) is ConditionalBreakNode:
            # at this point ConditionalBreakNode stores a claripy AST
            bool_var = src_block.condition
            if src_block.target == dst_block.addr:
                return bool_var
            else:
                return claripy.Not(bool_var)

        if type(src_block) is GraphRegion:
            return claripy.true

        # sometimes the last statement is the conditional jump. sometimes it's the first statement of the block
        if (
            isinstance(src_block, ailment.Block)
            and src_block.statements
            and isinstance(first_nonlabel_statement(src_block), ailment.Stmt.ConditionalJump)
        ):
            last_stmt = first_nonlabel_statement(src_block)
        else:
            last_stmt = self.get_last_statement(src_block)

        if last_stmt is None:
            return claripy.true
        if type(last_stmt) is ailment.Stmt.Jump:
            if isinstance(last_stmt.target, ailment.Expr.Const):
                return claripy.true
            # indirect jump
            target_ast = self.claripy_ast_from_ail_condition(last_stmt.target)
            return target_ast == dst_block.addr
        if type(last_stmt) is ailment.Stmt.ConditionalJump:
            bool_var = self.claripy_ast_from_ail_condition(last_stmt.condition)
            if isinstance(last_stmt.true_target, ailment.Expr.Const) and last_stmt.true_target.value == dst_block.addr:
                return bool_var
            else:
                return claripy.Not(bool_var)

        return claripy.true

    #
    # Expression conversion
    #

    def _convert_extract(self, hi, lo, expr, tags, memo=None):
        # ailment does not support Extract. We translate Extract to Convert and shift.
        if lo == 0:
            return ailment.Expr.Convert(
                None,
                expr.size(),
                hi + 1,
                False,
                self.convert_claripy_bool_ast(expr, memo=memo),
                **tags,
            )

        raise NotImplementedError("This case will be implemented once encountered.")

    def convert_claripy_bool_ast(self, cond, memo=None):
        """
        Convert recovered reaching conditions from claripy ASTs to ailment Expressions

        :return: None
        """

        if memo is None:
            memo = {}
        if cond._hash in memo:
            return memo[cond._hash]
        r = self.convert_claripy_bool_ast_core(cond, memo)
        optimized_r = peephole_optimize_expr(r, self._peephole_expr_optimizations)
        r = r if optimized_r is None else optimized_r
        memo[cond._hash] = r
        return r

    def convert_claripy_bool_ast_core(self, cond, memo):
        if isinstance(cond, ailment.Expr.Expression):
            return cond

        if cond.op in {"BoolS", "BoolV"} and claripy.is_true(cond):
            return ailment.Expr.Const(None, None, True, 1)
        if cond in self._condition_mapping:
            return self._condition_mapping[cond]
        if cond.op in {"BVS", "BoolS"} and cond.args[0] in self._condition_mapping:
            return self._condition_mapping[cond.args[0]]

        def _binary_op_reduce(op, args, tags, signed=False):
            r = None
            for arg in args:
                if r is None:
                    r = self.convert_claripy_bool_ast(arg, memo=memo)
                else:
                    r = ailment.Expr.BinaryOp(
                        None, op, (r, self.convert_claripy_bool_ast(arg, memo=memo)), signed, **tags
                    )
            return r

        def _unary_op_reduce(op, arg, tags):
            r = self.convert_claripy_bool_ast(arg, memo=memo)
            # TODO: Keep track of tags
            return ailment.Expr.UnaryOp(None, op, r, **tags)

        _mapping = {
            "Not": lambda cond_, tags: _unary_op_reduce("Not", cond_.args[0], tags),
            "__neg__": lambda cond_, tags: _unary_op_reduce("Not", cond_.args[0], tags),
            "__invert__": lambda cond_, tags: _unary_op_reduce("BitwiseNeg", cond_.args[0], tags),
            "And": lambda cond_, tags: _binary_op_reduce("LogicalAnd", cond_.args, tags),
            "Or": lambda cond_, tags: _binary_op_reduce("LogicalOr", cond_.args, tags),
            "__le__": lambda cond_, tags: _binary_op_reduce("CmpLE", cond_.args, tags, signed=True),
            "SLE": lambda cond_, tags: _binary_op_reduce("CmpLE", cond_.args, tags, signed=True),
            "__lt__": lambda cond_, tags: _binary_op_reduce("CmpLT", cond_.args, tags, signed=True),
            "SLT": lambda cond_, tags: _binary_op_reduce("CmpLT", cond_.args, tags, signed=True),
            "UGT": lambda cond_, tags: _binary_op_reduce("CmpGT", cond_.args, tags),
            "UGE": lambda cond_, tags: _binary_op_reduce("CmpGE", cond_.args, tags),
            "__gt__": lambda cond_, tags: _binary_op_reduce("CmpGT", cond_.args, tags, signed=True),
            "__ge__": lambda cond_, tags: _binary_op_reduce("CmpGE", cond_.args, tags, signed=True),
            "SGT": lambda cond_, tags: _binary_op_reduce("CmpGT", cond_.args, tags, signed=True),
            "SGE": lambda cond_, tags: _binary_op_reduce("CmpGE", cond_.args, tags, signed=True),
            "ULT": lambda cond_, tags: _binary_op_reduce("CmpLT", cond_.args, tags),
            "ULE": lambda cond_, tags: _binary_op_reduce("CmpLE", cond_.args, tags),
            "__eq__": lambda cond_, tags: _binary_op_reduce("CmpEQ", cond_.args, tags),
            "__ne__": lambda cond_, tags: _binary_op_reduce("CmpNE", cond_.args, tags),
            "__add__": lambda cond_, tags: _binary_op_reduce("Add", cond_.args, tags, signed=False),
            "__sub__": lambda cond_, tags: _binary_op_reduce("Sub", cond_.args, tags),
            "__mul__": lambda cond_, tags: _binary_op_reduce("Mul", cond_.args, tags),
            "__xor__": lambda cond_, tags: _binary_op_reduce("Xor", cond_.args, tags),
            "__or__": lambda cond_, tags: _binary_op_reduce("Or", cond_.args, tags, signed=False),
            "__and__": lambda cond_, tags: _binary_op_reduce("And", cond_.args, tags),
            "__lshift__": lambda cond_, tags: _binary_op_reduce("Shl", cond_.args, tags),
            "__rshift__": lambda cond_, tags: _binary_op_reduce("Sar", cond_.args, tags),
            "__floordiv__": lambda cond_, tags: _binary_op_reduce("Div", cond_.args, tags),
            "__mod__": lambda cond_, tags: _binary_op_reduce("Mod", cond_.args, tags),
            "LShR": lambda cond_, tags: _binary_op_reduce("Shr", cond_.args, tags),
            "BVV": lambda cond_, tags: ailment.Expr.Const(None, None, cond_.args[0], cond_.size(), **tags),
            "BoolV": lambda cond_, tags: ailment.Expr.Const(None, None, True, 1, **tags)
            if cond_.args[0] is True
            else ailment.Expr.Const(None, None, False, 1, **tags),
            "Extract": lambda cond_, tags: self._convert_extract(*cond_.args, tags, memo=memo),
        }

        if cond.op in _mapping:
            if cond in self._ast2annotations:
                cond_tags = self._ast2annotations.get(cond)
            elif claripy.Not(cond) in self._ast2annotations:
                cond_tags = self._ast2annotations.get(claripy.Not(cond))
            else:
                cond_tags = {}
            return _mapping[cond.op](cond, cond_tags)
        raise NotImplementedError(
            ("Condition variable %s has an unsupported operator %s. Consider implementing.") % (cond, cond.op)
        )

    def claripy_ast_from_ail_condition(self, condition, nobool: bool = False) -> claripy.ast.Bool:
        # Unpack a condition all the way to the leaves
        if isinstance(condition, claripy.ast.Base):  # pylint:disable=isinstance-second-argument-not-valid-type
            return condition

        if isinstance(
            condition,
            (ailment.Expr.DirtyExpression, ailment.Expr.BasePointerOffset, ailment.Expr.ITE),
        ):
            return _dummy_bvs(condition, self._condition_mapping)
        elif isinstance(condition, ailment.Stmt.Call):
            return _dummy_bvs(condition, self._condition_mapping, name_suffix=hex(condition.tags.get("ins_addr", 0)))
        elif isinstance(condition, (ailment.Expr.Load, ailment.Expr.Register)):
            # does it have a variable associated?
            if condition.variable is not None:
                var = claripy.BVS(
                    f"ailexpr_{repr(condition)}-{condition.variable.ident}", condition.bits, explicit_name=True
                )
            else:
                var = claripy.BVS(
                    "ailexpr_%s-%d" % (repr(condition), condition.idx), condition.bits, explicit_name=True
                )
            self._condition_mapping[var.args[0]] = condition
            return var
        elif isinstance(condition, ailment.Expr.Convert):
            # convert is special. if it generates a 1-bit variable, it should be treated as a BoolS
            if condition.to_bits == 1:
                var_ = self.claripy_ast_from_ail_condition(condition.operands[0])
                name = "ailcond_Conv(%d->%d, %d)" % (condition.from_bits, condition.to_bits, hash(var_))
                var = claripy.BoolS(name, explicit_name=True)
            else:
                var_ = self.claripy_ast_from_ail_condition(condition.operands[0])
                name = "ailexpr_Conv(%d->%d, %d)" % (condition.from_bits, condition.to_bits, hash(var_))
                var = claripy.BVS(name, condition.to_bits, explicit_name=True)
            self._condition_mapping[var.args[0]] = condition
            return var
        elif isinstance(condition, ailment.Expr.Const):
            if condition.value is True or condition.value is False:
                var = claripy.BoolV(condition.value)
            else:
                var = claripy.BVV(condition.value, condition.bits)
            return var
        elif isinstance(condition, ailment.Expr.Tmp):
            l.warning("Left-over ailment.Tmp variable %s.", condition)
            if condition.bits == 1:
                var = claripy.BoolS("ailtmp_%d" % condition.tmp_idx, explicit_name=True)
            else:
                var = claripy.BVS("ailtmp_%d" % condition.tmp_idx, condition.bits, explicit_name=True)
            self._condition_mapping[var.args[0]] = condition
            return var
        elif isinstance(condition, ailment.Expr.MultiStatementExpression):
            # just cache it
            if condition.bits == 1:
                var = claripy.BoolS("mstmtexpr_%d" % hash(condition), explicit_name=True)
            else:
                var = claripy.BVS("mstmtexpr_%d" % hash(condition), condition.bits, explicit_name=True)
            self._condition_mapping[var.args[0]] = condition
            return var

        lambda_expr = _ail2claripy_op_mapping.get(condition.verbose_op, None)
        if lambda_expr is None:
            # fall back to op
            lambda_expr = _ail2claripy_op_mapping.get(condition.op, None)
        if lambda_expr is None:
            raise NotImplementedError(
                "Unsupported AIL expression operation %s or %s. Consider implementing."
                % (condition.op, condition.verbose_op)
            )
        r = lambda_expr(condition, self.claripy_ast_from_ail_condition, self._condition_mapping)

        if isinstance(r, claripy.ast.Bool) and nobool:
            r = claripy.BVS("ailexpr_from_bool_%r" % r, 1, explicit_name=True)
            self._condition_mapping[r.args[0]] = condition

        if r is NotImplemented:
            if condition.bits == 1:
                r = claripy.BoolS("ailexpr_%r" % condition, explicit_name=True)
            else:
                r = claripy.BVS("ailexpr_%r" % condition, condition.bits, explicit_name=True)
            self._condition_mapping[r.args[0]] = condition
        # don't lose tags
        self._ast2annotations[r] = condition.tags
        return r

    #
    # Expression simplification
    #

    @staticmethod
    def claripy_ast_to_sympy_expr(ast, memo=None):
        if ast.op == "And":
            return sympy.And(*(ConditionProcessor.claripy_ast_to_sympy_expr(arg, memo=memo) for arg in ast.args))
        if ast.op == "Or":
            return sympy.Or(*(ConditionProcessor.claripy_ast_to_sympy_expr(arg, memo=memo) for arg in ast.args))
        if ast.op == "Not":
            return sympy.Not(ConditionProcessor.claripy_ast_to_sympy_expr(ast.args[0], memo=memo))

        if ast.op in _UNIFIABLE_COMPARISONS:
            # unify comparisons to enable more simplification opportunities without going "deep" in sympy
            inverse_op = getattr(ast.args[0], claripy.operations.inverse_operations[ast.op])
            return sympy.Not(ConditionProcessor.claripy_ast_to_sympy_expr(inverse_op(ast.args[1]), memo=memo))

        if memo is not None and ast in memo:
            return memo[ast]
        symbol = sympy.Symbol(str(hash(ast)))
        if memo is not None:
            memo[symbol] = ast
        return symbol

    @staticmethod
    def sympy_expr_to_claripy_ast(expr, memo: Dict):
        if expr.is_Symbol:
            return memo[expr]
        if isinstance(expr, sympy.Or):
            return claripy.Or(*(ConditionProcessor.sympy_expr_to_claripy_ast(arg, memo) for arg in expr.args))
        if isinstance(expr, sympy.And):
            return claripy.And(*(ConditionProcessor.sympy_expr_to_claripy_ast(arg, memo) for arg in expr.args))
        if isinstance(expr, sympy.Not):
            return claripy.Not(ConditionProcessor.sympy_expr_to_claripy_ast(expr.args[0], memo))
        if isinstance(expr, sympy.logic.boolalg.BooleanTrue):
            return claripy.true
        if isinstance(expr, sympy.logic.boolalg.BooleanFalse):
            return claripy.false
        raise RuntimeError("Unreachable reached")

    @staticmethod
    def simplify_condition(cond, depth_limit=8, variables_limit=8):
        memo = {}
        if cond.depth > depth_limit or len(cond.variables) > variables_limit:
            return cond
        sympy_expr = ConditionProcessor.claripy_ast_to_sympy_expr(cond, memo=memo)
        r = ConditionProcessor.sympy_expr_to_claripy_ast(sympy.simplify_logic(sympy_expr, deep=False), memo)
        return r

    @staticmethod
    def simplify_condition_deprecated(cond):
        # Z3's simplification may yield weird and unreadable results
        # hence we mostly rely on our own simplification. we only use Z3's simplification results when it returns a
        # concrete value.
        claripy_simplified = claripy.simplify(cond)
        if not claripy_simplified.symbolic:
            return claripy_simplified

        simplified = ConditionProcessor._fold_double_negations(cond)
        cond = simplified if simplified is not None else cond
        simplified = ConditionProcessor._revert_short_circuit_conditions(cond)
        cond = simplified if simplified is not None else cond
        simplified = ConditionProcessor._extract_common_subexpressions(cond)
        cond = simplified if simplified is not None else cond
        # simplified = ConditionProcessor._remove_redundant_terms(cond)
        # cond = simplified if simplified is not None else cond
        # in the end, use claripy's simplification to handle really easy cases again
        simplified = ConditionProcessor._simplify_trivial_cases(cond)
        cond = simplified if simplified is not None else cond
        return cond

    @staticmethod
    def _simplify_trivial_cases(cond):
        if cond.op == "And":
            new_args = []
            for arg in cond.args:
                claripy_simplified = claripy.simplify(arg)
                if claripy.is_true(claripy_simplified):
                    continue
                new_args.append(arg)
            return claripy.And(*new_args)

        return None

    @staticmethod
    def _revert_short_circuit_conditions(cond):
        # revert short-circuit conditions
        # !A||(A&&!B) ==> !(A&&B)

        if cond.op != "Or":
            return cond

        if len(cond.args) == 1:
            # redundant operator. get rid of it
            return cond.args[0]

        or_arg0, or_arg1 = cond.args[:2]
        if or_arg1.op == "And":
            pass
        elif or_arg0.op == "And":
            or_arg0, or_arg1 = or_arg1, or_arg0
        else:
            return cond

        not_a = or_arg0
        solver = claripy.SolverCacheless()

        if not_a.variables == or_arg1.args[0].variables:
            solver.add(not_a == or_arg1.args[0])
            not_b = or_arg1.args[1]
        elif not_a.variables == or_arg1.args[1].variables:
            solver.add(not_a == or_arg1.args[1])
            not_b = or_arg1.args[0]
        else:
            return cond

        if not solver.satisfiable():
            # found it!
            b = claripy.Not(not_b)
            a = claripy.Not(not_a)
            if len(cond.args) <= 2:
                return claripy.Not(claripy.And(a, b))
            else:
                return claripy.Or(claripy.Not(claripy.And(a, b)), *cond.args[2:])
        else:
            return cond

    @staticmethod
    def _fold_double_negations(cond):
        # !(!A) ==> A
        # !((!A) && (!B)) ==> A || B
        # !((!A) && B) ==> A || !B
        # !(A || B) ==> (!A && !B)

        if cond.op != "Not":
            return None
        if cond.args[0].op == "Not":
            return cond.args[0]

        if cond.args[0].op == "And" and len(cond.args[0].args) == 2:
            and_0, and_1 = cond.args[0].args
            if and_0.op == "Not" and and_1.op == "Not":
                expr = claripy.Or(and_0.args[0], and_1.args[0])
                return expr

            if and_0.op == "Not":  # and_1.op != "Not"
                expr = claripy.Or(and_0.args[0], ConditionProcessor.simplify_condition(claripy.Not(and_1)))
                return expr

        if cond.args[0].op == "Or" and len(cond.args[0].args) == 2:
            or_0, or_1 = cond.args[0].args
            expr = claripy.And(
                ConditionProcessor.simplify_condition(claripy.Not(or_0)),
                ConditionProcessor.simplify_condition(claripy.Not(or_1)),
            )
            return expr

        return None

    @staticmethod
    def _extract_common_subexpressions(cond):
        def _expr_inside_collection(expr_, coll_) -> bool:
            for ex_ in coll_:
                if expr_ is ex_:
                    return True
            return False

        # (A && B) || (A && C) => A && (B || C)
        if cond.op == "And":
            args = [ConditionProcessor._extract_common_subexpressions(arg) for arg in cond.args]
            if all(arg is None for arg in args):
                return None
            return claripy.And(*((arg if arg is not None else ori_arg) for arg, ori_arg in zip(args, cond.args)))

        if cond.op == "Or":
            args = [ConditionProcessor._extract_common_subexpressions(arg) for arg in cond.args]
            args = [(arg if arg is not None else ori_arg) for arg, ori_arg in zip(args, cond.args)]

            expr_ctrs = defaultdict(int)
            for arg in args:
                if arg.op == "And":
                    for subexpr in arg.args:
                        expr_ctrs[subexpr] += 1
                else:
                    expr_ctrs[arg] += 1

            common_exprs = []
            for expr, ctr in expr_ctrs.items():
                if ctr == len(args):
                    # found a common one
                    common_exprs.append(expr)

            if not common_exprs:
                return claripy.Or(*args)

            new_args = []
            for arg in args:
                if arg.op == "And":
                    new_subexprs = [
                        subexpr for subexpr in arg.args if not _expr_inside_collection(subexpr, common_exprs)
                    ]
                    new_args.append(claripy.And(*new_subexprs))
                elif arg in common_exprs:
                    continue
                else:
                    raise RuntimeError("Unexpected behavior - you should never reach here")

            return claripy.And(*common_exprs, claripy.Or(*new_args))

        return None

    @staticmethod
    def _extract_terms(ast: claripy.ast.Bool) -> Generator[claripy.ast.Bool, None, None]:
        if ast.op == "And":
            for arg in ast.args:
                yield from ConditionProcessor._extract_terms(arg)
        elif ast.op == "Or":
            for arg in ast.args:
                yield from ConditionProcessor._extract_terms(arg)
        elif ast.op == "Not":
            yield from ConditionProcessor._extract_terms(ast.args[0])
        else:
            yield ast

    @staticmethod
    def _replace_term_in_ast(
        ast: claripy.ast.Bool,
        r0: claripy.ast.Bool,
        r0_with: claripy.ast.Bool,
        r1: claripy.ast.Bool,
        r1_with: claripy.ast.Bool,
    ) -> claripy.ast.Bool:
        if ast.op == "And":
            return ast.make_like(
                "And", (ConditionProcessor._replace_term_in_ast(arg, r0, r0_with, r1, r1_with) for arg in ast.args)
            )
        elif ast.op == "Or":
            return ast.make_like(
                "Or", (ConditionProcessor._replace_term_in_ast(arg, r0, r0_with, r1, r1_with) for arg in ast.args)
            )
        elif ast.op == "Not":
            return ast.make_like(
                "Not", (ConditionProcessor._replace_term_in_ast(ast.args[0], r0, r0_with, r1, r1_with),)
            )
        else:
            if ast is r0:
                return r0_with
            if ast is r1:
                return r1_with
            return ast

    @staticmethod
    def _remove_redundant_terms(cond):
        """
        Extract all terms and test for each term if its truism impacts the truism of the entire condition. If not, the
        term is redundant and can be replaced with a True.
        """

        all_terms = set()
        for term in ConditionProcessor._extract_terms(cond):
            if term not in all_terms:
                all_terms.add(term)

        negations = {}
        to_skip = set()
        all_terms_without_negs = set()
        for term in all_terms:
            if term in to_skip:
                continue
            neg = claripy.Not(term)
            if neg in all_terms:
                negations[term] = neg
                to_skip.add(neg)
                all_terms_without_negs.add(term)
            else:
                all_terms_without_negs.add(term)

        solver = claripy.SolverCacheless()
        for term in all_terms_without_negs:
            neg = negations.get(term, None)

            replaced_with_true = ConditionProcessor._replace_term_in_ast(cond, term, claripy.true, neg, claripy.false)
            sat0 = solver.satisfiable(
                extra_constraints=(
                    cond,
                    claripy.Not(replaced_with_true),
                )
            )
            sat1 = solver.satisfiable(
                extra_constraints=(
                    claripy.Not(cond),
                    replaced_with_true,
                )
            )
            if sat0 or sat1:
                continue

            replaced_with_false = ConditionProcessor._replace_term_in_ast(cond, term, claripy.false, neg, claripy.true)
            sat0 = solver.satisfiable(
                extra_constraints=(
                    cond,
                    claripy.Not(replaced_with_false),
                )
            )
            sat1 = solver.satisfiable(
                extra_constraints=(
                    claripy.Not(cond),
                    replaced_with_false,
                )
            )
            if sat0 or sat1:
                continue

            # TODO: Finish the implementation
            print(term, "is redundant")

    #
    # Graph processing
    #

    @staticmethod
    def _remove_crossing_edges_between_cases(
        graph: networkx.DiGraph, case_entry_to_switch_head: Dict[int, int]
    ) -> networkx.DiGraph:
        starting_nodes = {node for node in graph if node.addr in case_entry_to_switch_head}
        if not starting_nodes:
            return graph

        traversed_nodes = set()
        edges_to_remove = set()
        for starting_node in starting_nodes:
            queue = [starting_node]
            while queue:
                src = queue.pop(0)
                traversed_nodes.add(src)
                successors = graph.successors(src)
                for succ in successors:
                    if succ in traversed_nodes:
                        # we should not traverse this node twice
                        if graph.out_degree(succ) > 0:
                            edges_to_remove.add((src, succ))
                        continue
                    if succ in starting_nodes:
                        # we do not want any jump from one node to a starting node
                        edges_to_remove.add((src, succ))
                        continue
                    traversed_nodes.add(src)
                    queue.append(succ)

        if not edges_to_remove:
            return graph

        # make a copy before modifying the graph
        graph = networkx.DiGraph(graph)
        graph.remove_edges_from(edges_to_remove)
        return graph

    #
    # Utils
    #

    def create_jump_target_var(self, jumptable_head_addr: int):
        return claripy.BVS("jump_table_%x" % jumptable_head_addr, self.arch.bits, explicit_name=True)



from math import gcd

from ailment.expression import BinaryOp, UnaryOp, Const

from .base import PeepholeOptimizationExprBase


class EagerEvaluation(PeepholeOptimizationExprBase):
    """
    Eagerly evaluates certain types of expressions.
    """

    __slots__ = ()

    NAME = "Eager expression evaluation"
    expr_classes = (BinaryOp, UnaryOp)

    def optimize(self, expr, **kwargs):
        if isinstance(expr, BinaryOp):
            return self._optimize_binaryop(expr)
        elif isinstance(expr, UnaryOp):
            return self._optimize_unaryop(expr)
        return None

    @staticmethod
    def _optimize_binaryop(expr: BinaryOp):
        if expr.op == "Add":
            if isinstance(expr.operands[0], Const) and isinstance(expr.operands[1], Const):
                mask = (1 << expr.bits) - 1
                new_expr = Const(
                    expr.idx, None, (expr.operands[0].value + expr.operands[1].value) & mask, expr.bits, **expr.tags
                )
                return new_expr
            if isinstance(expr.operands[1], Const) and expr.operands[1].value == 0:
                return expr.operands[0]
            if (
                isinstance(expr.operands[1], Const)
                and isinstance(expr.operands[0], BinaryOp)
                and isinstance(expr.operands[0].operands[1], Const)
            ):
                left = expr.operands[0]
                inner_expr, const_0 = left.operands
                const_1 = expr.operands[1]
                if left.op == "Add":
                    return BinaryOp(
                        left.idx,
                        "Add",
                        [inner_expr, Const(const_0.idx, None, const_0.value + const_1.value, const_0.bits)],
                        expr.signed,
                        **expr.tags,
                    )
                elif left.op == "Sub":
                    return BinaryOp(
                        left.idx,
                        "Add",
                        [inner_expr, Const(const_0.idx, None, const_1.value - const_0.value, const_0.bits)],
                        expr.signed,
                        **expr.tags,
                    )
        elif expr.op == "Sub":
            if isinstance(expr.operands[0], Const) and isinstance(expr.operands[1], Const):
                mask = (1 << expr.bits) - 1
                new_expr = Const(
                    expr.idx, None, (expr.operands[0].value - expr.operands[1].value) & mask, expr.bits, **expr.tags
                )
                return new_expr
            if (
                isinstance(expr.operands[1], Const)
                and isinstance(expr.operands[0], BinaryOp)
                and isinstance(expr.operands[0].operands[1], Const)
            ):
                left = expr.operands[0]
                inner_expr, const_0 = left.operands
                const_1 = expr.operands[1]
                if left.op == "Add":
                    return BinaryOp(
                        left.idx,
                        "Sub",
                        [inner_expr, Const(const_0.idx, None, const_1.value - const_0.value, const_0.bits)],
                        expr.signed,
                        **expr.tags,
                    )
                elif left.op == "Sub":
                    return BinaryOp(
                        left.idx,
                        "Sub",
                        [inner_expr, Const(const_0.idx, None, const_0.value + const_1.value, const_0.bits)],
                        expr.signed,
                        **expr.tags,
                    )
            if isinstance(expr.operands[0], Const) and expr.operands[0].value == 0:
                return UnaryOp(expr.idx, "Neg", expr.operands[1], **expr.tags)

        elif expr.op == "And":
            if isinstance(expr.operands[0], Const) and isinstance(expr.operands[1], Const):
                new_expr = Const(
                    expr.idx, None, (expr.operands[0].value & expr.operands[1].value), expr.bits, **expr.tags
                )
                return new_expr
            if isinstance(expr.operands[1], Const) and expr.operands[1].value == 0:
                return Const(expr.idx, None, 0, expr.bits, **expr.tags)

        elif expr.op == "Mul" and isinstance(expr.operands[1], Const) and expr.operands[1].value == 1:
            return expr.operands[0]

        elif (
            expr.op == "Div"
            and isinstance(expr.operands[1], Const)
            and isinstance(expr.operands[0], BinaryOp)
            and expr.operands[0].op == "Mul"
            and isinstance(expr.operands[0].operands[1], Const)
        ):
            expr0, const_0 = expr.operands
            const_1 = expr0.operands[1]
            if const_0.value != 0 and const_1.value != 0:
                gcd_ = gcd(const_0.value, const_1.value)
                if gcd_ != 1:
                    new_const_1 = Const(
                        const_1.idx, const_1.variable, const_1.value // gcd_, const_1.bits, **const_1.tags
                    )
                    new_const_0 = Const(
                        const_0.idx, const_0.variable, const_0.value // gcd_, const_0.bits, **const_0.tags
                    )
                    mul = BinaryOp(
                        expr0.idx,
                        "Mul",
                        (expr0.operands[0], new_const_1),
                        expr0.signed,
                        variable=expr0.variable,
                        variable_offset=expr0.variable_offset,
                        bits=expr0.bits,
                        **expr0.tags,
                    )
                    div = BinaryOp(expr.idx, "Div", (mul, new_const_0), expr.signed, bits=expr.bits, **expr.tags)
                    return div

        elif expr.op in {"Shr", "Sar"} and isinstance(expr.operands[1], Const):
            expr0, expr1 = expr.operands
            if isinstance(expr0, BinaryOp) and expr0.op == "Shr" and isinstance(expr0.operands[1], Const):
                # (a >> M) >> N  ==>  a >> (M + N)
                const_a = expr0.operands[1]
                const_b = expr1
                const = Const(const_b.idx, None, const_a.value + const_b.value, const_b.bits, **const_b.tags)
                new_expr = BinaryOp(expr.idx, expr.op, (expr0.operands[0], const), False, bits=expr.bits, **expr.tags)
                return new_expr

            if isinstance(expr0, BinaryOp) and expr0.op == "Div" and isinstance(expr0.operands[1], Const):
                # (a / M0) >> M1  ==>  a / (M0 * 2 ** M1)
                const_m0 = expr0.operands[1]
                const_m1 = expr1
                const = Const(const_m0.idx, None, const_m0.value * 2**const_m1.value, const_m0.bits, **const_m0.tags)
                new_expr = BinaryOp(
                    expr.idx, "Div", (expr0.operands[0], const), expr.signed, bits=expr.bits, **expr.tags
                )
                return new_expr

            if isinstance(expr0, Const):
                const_a = expr0.value
                mask = (2**expr0.bits) - 1
                new_expr = Const(expr0.idx, None, (const_a >> expr1.value) & mask, expr0.bits, **expr0.tags)
                return new_expr

            if expr.op == "Shr" and expr.operands[0].bits <= expr.operands[1].value:
                return Const(expr.idx, None, 0, expr.operands[0].bits, **expr.tags)

        elif expr.op == "Shl" and isinstance(expr.operands[1], Const):
            expr0, expr1 = expr.operands
            if isinstance(expr0, Const):
                const_a = expr0.value
                mask = (2**expr0.bits) - 1
                new_expr = Const(expr0.idx, None, (const_a << expr1.value) & mask, expr0.bits, **expr0.tags)
                return new_expr

        elif expr.op == "Or":
            if isinstance(expr.operands[0], Const) and isinstance(expr.operands[1], Const):
                return Const(expr.idx, None, expr.operands[0].value | expr.operands[1].value, expr.bits, **expr.tags)
            if isinstance(expr.operands[0], Const) and expr.operands[0].value == 0:
                return expr.operands[1]
            if isinstance(expr.operands[1], Const) and expr.operands[1].value == 0:
                return expr.operands[0]
            if expr.operands[0].likes(expr.operands[1]):
                return expr.operands[0]

        return None

    @staticmethod
    def _optimize_unaryop(expr: UnaryOp):
        if expr.op == "Neg" and isinstance(expr.operand, Const):
            const_a = expr.operand.value
            mask = (2**expr.bits) - 1
            new_expr = Const(expr.idx, None, (~const_a) & mask, expr.bits, **expr.tags)
            return new_expr

        return None



#!/usr/bin/env python3
# pylint:disable=missing-class-docstring,no-self-use
__package__ = __package__ or "tests.sim.options"  # pylint:disable=redefined-builtin

import os
import unittest

import angr

from ...common import bin_location


test_location = os.path.join(bin_location, "tests")


def find(s):
    return s.posix.dumps(1) == b"c0de\n"


def avoid(s):
    return s.posix.dumps(1) == b"nothing\n"


class TestSymbolicFd(unittest.TestCase):
    def test_symbolic_fd(self):
        project = angr.Project(os.path.join(test_location, "x86_64", "symbolic_fd"))

        for method_name in ("stat_test", "fstat_test", "open_test", "fopen_test", "fdopen_test"):
            addr = project.loader.find_symbol(method_name).rebased_addr

            # all files exist
            state = project.factory.blank_state(addr=addr)
            state.options["ALL_FILES_EXIST"] = True
            simgr = project.factory.simgr(state)
            while simgr.active != []:
                simgr.explore(find=find, avoid=avoid)
            assert simgr.avoid != [] and simgr.found == [], f"{method_name}: got {simgr.avoid} and {simgr.found}"

            # any file might exist
            state = project.factory.blank_state(addr=addr)
            state.options["ALL_FILES_EXIST"] = False
            state.options["ANY_FILE_MIGHT_EXIST"] = True
            simgr = project.factory.simgr(state)
            while simgr.active != []:
                simgr.explore(find=find, avoid=avoid)
            assert simgr.avoid != [] and simgr.found != [], f"{method_name}: got {simgr.avoid} and {simgr.found}"

            # no file exists
            state = project.factory.blank_state(addr=addr)
            state.options["ALL_FILES_EXIST"] = False
            state.options["ANY_FILE_MIGHT_EXIST"] = False
            simgr = project.factory.simgr(state)
            while simgr.active != []:
                simgr.explore(find=find, avoid=avoid)
            assert simgr.avoid == [] and simgr.found != [], f"{method_name}: got {simgr.avoid} and {simgr.found}"


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# pylint: disable=missing-class-docstring,disable=no-self-use
__package__ = __package__ or "tests.serialization"  # pylint:disable=redefined-builtin

import os
import tempfile
import unittest

import angr
from angr.angrdb import AngrDB

from ..common import bin_location


test_location = os.path.join(bin_location, "tests")


class TestDb(unittest.TestCase):
    def test_angrdb_fauxware(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")

        proj = angr.Project(bin_path, auto_load_libs=False)
        cfg: angr.analyses.CFGFast = proj.analyses.CFGFast(data_references=True, cross_references=True, normalize=True)
        proj.kb.comments[proj.entry] = "Entry point"

        dtemp = tempfile.mkdtemp()
        db_file = os.path.join(dtemp, "fauxware.adb")

        db = AngrDB(proj)
        db.dump(db_file)

        db1 = AngrDB()
        new_proj = db1.load(db_file)

        assert len(new_proj.kb.cfgs["CFGFast"].nodes()) == len(cfg.model.nodes())
        assert len(new_proj.kb.functions) == len(proj.kb.functions)

        # compare each function
        for func in proj.kb.functions.values():
            new_func = new_proj.kb.functions[func.addr]

            assert func.addr == new_func.addr
            assert func.normalized == new_func.normalized

            assert len(func.transition_graph.nodes()) == len(new_func.transition_graph.nodes())
            assert set(map(lambda x: x.addr, func.transition_graph.nodes())) == set(
                map(lambda x: x.addr, new_func.transition_graph.nodes())
            )
            assert len(func.transition_graph.edges()) == len(new_func.transition_graph.edges())

        # compare call graph
        callgraph_nodes_old = set(proj.kb.callgraph.nodes)
        callgraph_nodes_new = set(new_proj.kb.callgraph.nodes)
        callgraph_edges_old = set(proj.kb.callgraph.edges)
        callgraph_edges_new = set(new_proj.kb.callgraph.edges)

        assert callgraph_nodes_old == callgraph_nodes_new
        assert callgraph_edges_old == callgraph_edges_new

        # compare CFG
        new_cfg = new_proj.kb.cfgs["CFGFast"]
        for node in cfg.model.nodes():
            new_node = new_cfg.get_any_node(node.addr)

            assert new_node.addr == node.addr
            assert new_node.size == node.size

        # compare memory data
        for addr, memory_data in cfg.model.memory_data.items():
            new_memory_data = new_cfg.memory_data[addr]

            assert memory_data.addr == new_memory_data.addr
            assert memory_data.size == new_memory_data.size
            assert memory_data.reference_size == new_memory_data.reference_size
            assert memory_data.sort == new_memory_data.sort
            assert memory_data.content == new_memory_data.content

        assert cfg.model.insn_addr_to_memory_data.keys() == new_cfg.insn_addr_to_memory_data.keys()

        # comments
        for addr, comment in proj.kb.comments.items():
            new_comment = new_proj.kb.comments.get(addr, None)

            assert comment == new_comment

    def test_angrdb_open_multiple_times(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")

        proj = angr.Project(bin_path, auto_load_libs=False)
        _: angr.analyses.CFGFast = proj.analyses.CFGFast(data_references=True, cross_references=True, normalize=True)
        proj.kb.comments[proj.entry] = "Entry point"

        dtemp = tempfile.mkdtemp()
        db_file = os.path.join(dtemp, "fauxware.adb")

        db = AngrDB(proj)
        db.dump(db_file)

        # attempt 0
        db0 = AngrDB()
        proj0 = db0.load(db_file)

        # attempt 1
        db1 = AngrDB()
        proj1 = db1.load(db_file)

        # attempt 2
        db2 = AngrDB()
        proj2 = db2.load(db_file)

        # attempt 3
        db3 = AngrDB()
        proj3 = db3.load(db_file)

        # compare functions
        for func in proj.kb.functions.values():
            for p in [proj0, proj1, proj2, proj3]:
                new_func = p.kb.functions[func.addr]

                assert func.addr == new_func.addr
                assert func.normalized == new_func.normalized

                assert len(func.transition_graph.nodes()) == len(new_func.transition_graph.nodes())
                assert set(map(lambda x: x.addr, func.transition_graph.nodes())) == set(
                    map(lambda x: x.addr, new_func.transition_graph.nodes())
                )
                assert len(func.transition_graph.edges()) == len(new_func.transition_graph.edges())

    def test_angrdb_save_multiple_times(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")

        proj = angr.Project(bin_path, auto_load_libs=False)
        _: angr.analyses.CFGFast = proj.analyses.CFGFast(data_references=True, cross_references=True, normalize=True)
        proj.kb.comments[proj.entry] = "Entry point"

        dtemp = tempfile.mkdtemp()
        db_file = os.path.join(dtemp, "fauxware.adb")

        # attempt 0
        db = AngrDB(proj)
        db.dump(db_file)

        # attempt 1
        proj0 = AngrDB().load(db_file)
        assert proj0.kb.comments[proj.entry] == "Entry point"
        proj0.kb.comments[proj.entry] = "Comment 0"
        AngrDB(proj0).dump(db_file)

        # attempt 2
        proj1 = AngrDB().load(db_file)
        assert proj1.kb.comments[proj.entry] == "Comment 0"
        proj1.kb.comments[proj.entry] = "Comment 1"
        AngrDB(proj1).dump(db_file)

        # attempt 3
        proj1 = AngrDB().load(db_file)
        assert proj1.kb.comments[proj.entry] == "Comment 1"
        proj1.kb.comments[proj.entry] = "Comment 22222222222222222222222"
        AngrDB(proj1).dump(db_file)

        # attempt 4
        proj1 = AngrDB().load(db_file)
        assert proj1.kb.comments[proj.entry] == "Comment 22222222222222222222222"


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# pylint: disable=missing-class-docstring,no-self-use,line-too-long

import time
import unittest

import claripy

from angr.storage.memory_mixins import (
    DataNormalizationMixin,
    SizeNormalizationMixin,
    AddressConcretizationMixin,
    UltraPagesMixin,
    ListPagesMixin,
    PagedMemoryMixin,
    MultiValuedMemory,
    MVListPagesMixin,
)
from angr.storage.memory_mixins.paged_memory.pages.multi_values import MultiValues
from angr import SimState, SIM_PROCEDURES
from angr import options as o
from angr.state_plugins import SimSystemPosix, SimLightRegisters
from angr.storage.file import SimFile


class UltraPageMemory(
    DataNormalizationMixin,
    SizeNormalizationMixin,
    AddressConcretizationMixin,
    UltraPagesMixin,
    PagedMemoryMixin,
):
    pass


class ListPageMemory(
    DataNormalizationMixin,
    SizeNormalizationMixin,
    AddressConcretizationMixin,
    ListPagesMixin,
    PagedMemoryMixin,
):
    pass


class MVPageMemory(
    SizeNormalizationMixin,
    AddressConcretizationMixin,
    MVListPagesMixin,
    PagedMemoryMixin,
):
    pass


class TestMemory(unittest.TestCase):
    def test_copy(self):
        s = SimState(arch="AMD64", mode="symbolic")
        s.memory.store(0x100, b"ABCDEFGHIJKLMNOP")
        s.memory.store(0x200, b"XXXXXXXXXXXXXXXX")
        x = s.solver.BVS("size", s.arch.bits)
        s.add_constraints(s.solver.ULT(x, 10))
        s.memory.copy_contents(0x200, 0x100, x)

        assert sorted(s.solver.eval_upto(x, 100)) == list(range(10))
        result = s.memory.load(0x200, 5)
        assert sorted(s.solver.eval_upto(result, 100, cast_to=bytes)) == [
            b"ABCDE",
            b"ABCDX",
            b"ABCXX",
            b"ABXXX",
            b"AXXXX",
            b"XXXXX",
        ]
        assert sorted(s.solver.eval_upto(result, 100, cast_to=bytes, extra_constraints=[x == 3])) == [b"ABCXX"]

        s = SimState(arch="AMD64", mode="symbolic")
        s.register_plugin(
            "posix", SimSystemPosix(stdin=SimFile(name="stdin", content=b"ABCDEFGHIJKLMNOP", has_end=True))
        )
        s.memory.store(0x200, b"XXXXXXXXXXXXXXXX")
        x = s.solver.BVS("size", s.arch.bits)
        s.add_constraints(s.solver.ULT(x, 10))

        s.posix.get_fd(0).read(0x200, x)
        assert sorted(s.solver.eval_upto(x, 100)) == list(range(10))
        result = s.memory.load(0x200, 5)
        assert sorted(s.solver.eval_upto(result, 100, cast_to=bytes)) == [
            b"ABCDE",
            b"ABCDX",
            b"ABCXX",
            b"ABXXX",
            b"AXXXX",
            b"XXXXX",
        ]
        assert sorted(s.solver.eval_upto(result, 100, cast_to=bytes, extra_constraints=[x == 3])) == [b"ABCXX"]

        s = SimState(arch="AMD64", mode="symbolic")
        s.register_plugin("posix", SimSystemPosix(stdin=SimFile(name="stdin", content=b"ABCDEFGHIJKLMNOP")))
        s.memory.store(0x200, b"XXXXXXXXXXXXXXXX")
        x = s.solver.BVS("size", s.arch.bits)
        s.add_constraints(s.solver.ULT(x, 10))

        read_proc = SIM_PROCEDURES["posix"]["read"]()
        ret_x = read_proc.execute(s, arguments=(0, 0x200, x)).ret_expr
        assert sorted(s.solver.eval_upto(x, 100)) == list(range(10))
        result = s.memory.load(0x200, 5)
        assert sorted(s.solver.eval_upto(result, 100, cast_to=bytes)) == [
            b"ABCDE",
            b"ABCDX",
            b"ABCXX",
            b"ABXXX",
            b"AXXXX",
            b"XXXXX",
        ]
        assert sorted(s.solver.eval_upto(result, 100, cast_to=bytes, extra_constraints=[x == 3])) == [b"ABCXX"]

        assert sorted(s.solver.eval_upto(ret_x, 100)) == list(range(10))
        assert sorted(s.solver.eval_upto(result, 100, cast_to=bytes, extra_constraints=[ret_x == 3])) == [b"ABCXX"]

    @staticmethod
    def _concrete_memory_tests(s):
        # Store a 4-byte variable to memory directly...
        s.memory.store(100, s.solver.BVV(0x1337, 32))
        # ... then load it

        expr = s.memory.load(100, 4)
        assert expr is s.solver.BVV(0x1337, 32)
        expr = s.memory.load(100, 2)
        assert expr is s.solver.BVV(0, 16)
        expr = s.memory.load(102, 2)
        assert expr is s.solver.BVV(0x1337, 16)

        # partially symbolic
        expr = s.memory.load(102, 4)
        assert expr.length == 32
        assert s.solver.min(expr) == 0x13370000
        assert s.solver.max(expr) == 0x1337FFFF

        # partial overwrite
        s.memory.store(101, s.solver.BVV(0x1415, 16))
        expr = s.memory.load(101, 3)
        assert expr is s.solver.BVV(0x141537, 24)
        expr = s.memory.load(100, 2)
        assert s.solver.min(expr) == 0x14
        expr = s.memory.load(102, 2)
        assert expr is s.solver.BVV(0x1537, 16)
        expr = s.memory.load(102, 2, endness="Iend_LE")
        assert expr is s.solver.BVV(0x3715, 16)

        s.memory.store(0x100, s.solver.BVV(b"AAAABBBBCCCCDDDDEEEEFFFFGGGGHHHH"), endness="Iend_LE")
        expr = s.memory.load(0x104, 13)
        assert expr is s.solver.BVV(b"GGGGFFFFEEEED")

        # branching
        s2 = s.copy()
        s2a = s2.copy()
        s2b = s2.copy()

        s2a.memory.store(0x100, s.solver.BVV(b"A"))
        s2b.memory.store(0x100, s.solver.BVV(b"B"))
        assert s2b.memory.load(0x100, 1) is s.solver.BVV(b"B")
        assert s2a.memory.load(0x100, 1) is s.solver.BVV(b"A")

    ## pylint: disable=R0904
    def test_memory(self):
        initial_memory = {0: b"A", 1: b"A", 2: b"A", 3: b"A", 10: b"B"}
        s = SimState(
            arch="AMD64",
            dict_memory_backer=initial_memory,
            add_options={o.REVERSE_MEMORY_NAME_MAP, o.REVERSE_MEMORY_HASH_MAP},
        )

        self._concrete_memory_tests(s)
        # concrete address and partially symbolic result
        expr = s.memory.load(2, 4)
        expr = s.memory.load(2, 4)
        expr = s.memory.load(2, 4)
        expr = s.memory.load(2, 4)
        assert s.solver.symbolic(expr)
        assert not s.solver.unique(expr)
        assert s.solver.eval(expr) >= 0x41410000
        assert s.solver.eval(expr) <= 0x41420000
        assert s.solver.min_int(expr) == 0x41410000
        assert s.solver.max_int(expr) == 0x4141FFFF

        # concrete address and concrete result
        expr = s.memory.load(0, 4)  # Returns: a z3 BVS representing 0x41414141
        assert not s.solver.symbolic(expr)
        assert s.solver.eval(expr) == 0x41414141

        c = s.solver.BVS("condition", 8)
        expr = s.memory.load(10, 1, condition=c == 1, fallback=s.solver.BVV(b"X"))
        assert s.solver.eval_upto(expr, 10, cast_to=bytes, extra_constraints=[c == 1]) == [b"B"]
        assert s.solver.eval_upto(expr, 10, cast_to=bytes, extra_constraints=[c != 1]) == [b"X"]

        x = s.solver.BVS("ref_test", 16, explicit_name=True)
        s.memory.store(0x1000, x)
        s.memory.store(0x2000, x)
        assert set(s.memory.addrs_for_name("ref_test")) == {0x1000, 0x1001, 0x2000, 0x2001}
        assert set(s.memory.addrs_for_hash(x.cache_key)) == {0x1000, 0x1001, 0x2000, 0x2001}

        s2 = s.copy()
        y = s2.solver.BVS("ref_test2", 16, explicit_name=True)
        s2.memory.store(0x2000, y)
        assert s2.memory.load(0x2000, 2) is y
        assert s.memory.load(0x2000, 2) is x
        assert set(s.memory.addrs_for_name("ref_test")) == {0x1000, 0x1001, 0x2000, 0x2001}
        assert set(s.memory.addrs_for_hash(x.cache_key)) == {0x1000, 0x1001, 0x2000, 0x2001}
        assert set(s2.memory.addrs_for_name("ref_test")) == {0x1000, 0x1001}
        assert set(s2.memory.addrs_for_hash(x.cache_key)) == {0x1000, 0x1001}
        assert set(s2.memory.addrs_for_name("ref_test2")) == {0x2000, 0x2001}
        assert set(s2.memory.addrs_for_hash(y.cache_key)) == {0x2000, 0x2001}

        s.memory.store(0x3000, s.solver.BVS("replace_old", 32, explicit_name=True))
        s.memory.store(0x3001, s.solver.BVV(b"AB"))
        assert set(s.memory.addrs_for_name("replace_old")) == {0x3000, 0x3003}
        assert s.solver.eval_upto(s.memory.load(0x3001, 2), 10, cast_to=bytes) == [b"AB"]

        # n = s.solver.BVS('replace_new', 32, explicit_name=True)
        # c = s.solver.BVS('replace_cool', 32, explicit_name=True)

        # mo = s.memory.memory_objects_for_name('replace_old')
        # assert len(mo) == 1
        # s.memory.replace_memory_object(next(iter(mo)), n)
        # assert set(s.memory.addrs_for_name('replace_old')) == set()
        # assert set(s.memory.addrs_for_name('replace_new')) == {0x3000, 0x3003}
        # assert s.solver.eval_upto(s.memory.load(0x3001, 2), 10, cast_to=bytes) == [b"AB"]

        # s.memory.store(0x4000, s.solver.If(n == 0, n+10, n+20))

        # assert set(s.memory.addrs_for_name('replace_new')) == {0x3000, 0x3003, 0x4000, 0x4001, 0x4002, 0x4003}
        # s.memory.replace_all(n, c)
        # assert set(s.memory.addrs_for_name('replace_old')) == set()
        # assert set(s.memory.addrs_for_name('replace_new')) == set()
        # assert set(s.memory.addrs_for_name('replace_cool')) == {0x3000, 0x3003, 0x4000, 0x4001, 0x4002, 0x4003}
        # assert s.solver.eval_upto(s.memory.load(0x3001, 2), 10, cast_to=bytes) == [b"AB"]

        # z = s.solver.BVV(0, 32)
        # s.memory.replace_all(c, z)
        # assert set(s.memory.addrs_for_name('replace_old')) == set()
        # assert set(s.memory.addrs_for_name('replace_new')) == set()
        # assert set(s.memory.addrs_for_name('replace_cool')) == set()
        # assert s.solver.eval_upto(s.memory.load(0x3001, 2), 10, cast_to=bytes) == [b"AB"]
        # assert s.solver.eval_upto(s.memory.load(0x3000, 4), 10) == [0x00414200]
        # assert s.solver.eval_upto(s.memory.load(0x4000, 4), 10) == [0x0000000a]

        # symbolic length
        x = s.solver.BVV(0x11223344, 32)
        y = s.solver.BVV(0xAABBCCDD, 32)
        n = s.solver.BVS("size", 32)
        s.add_constraints(n <= 4)
        s.memory.store(0x5000, x)
        s.memory.store(0x5000, y, size=n)
        assert set(s.solver.eval_upto(s.memory.load(0x5000, 4), 10)) == {
            0x11223344,
            0xAA223344,
            0xAABB3344,
            0xAABBCC44,
            0xAABBCCDD,
        }

        s1 = s.copy()
        s1.add_constraints(n == 1)
        assert set(s1.solver.eval_upto(s1.memory.load(0x5000, 4), 10)) == {0xAA223344}

        s4 = s.copy()
        s4.add_constraints(n == 4)
        assert set(s4.solver.eval_upto(s4.memory.load(0x5000, 4), 10)) == {0xAABBCCDD}

        # condition without fallback
        x = s.solver.BVV(0x11223344, 32)
        y = s.solver.BVV(0xAABBCCDD, 32)
        c = s.solver.BVS("condition", 32)
        s.memory.store(0x6000, x)
        s.memory.store(0x6000, y, condition=c == 1)
        assert set(s.solver.eval_upto(s.memory.load(0x6000, 4), 10)) == {0x11223344, 0xAABBCCDD}

        s0 = s.copy()
        s0.add_constraints(c == 0)
        assert set(s0.solver.eval_upto(s0.memory.load(0x6000, 4), 10)) == {0x11223344}

        s1 = s.copy()
        s1.add_constraints(c == 1)
        assert set(s1.solver.eval_upto(s1.memory.load(0x6000, 4), 10)) == {0xAABBCCDD}

        # condition with symbolic size
        x = s.solver.BVV(0x11223344, 32)
        y = s.solver.BVV(0xAABBCCDD, 32)
        c = s.solver.BVS("condition", 32)
        n = s.solver.BVS("size", 32)
        s.add_constraints(n <= 4)
        s.memory.store(0x8000, x)
        s.memory.store(0x8000, y, condition=c == 1, size=n)

        s0 = s.copy()
        s0.add_constraints(c == 0)
        assert set(s0.solver.eval_upto(s0.memory.load(0x8000, 4), 10)) == {0x11223344}

        s1 = s.copy()
        s1.add_constraints(c == 1)
        assert set(s1.solver.eval_upto(s1.memory.load(0x8000, 4), 10)) == {
            0x11223344,
            0xAA223344,
            0xAABB3344,
            0xAABBCC44,
            0xAABBCCDD,
        }

    def test_abstract_memory(self):
        initial_memory = {0: b"A", 1: b"B", 2: b"C", 3: b"D"}

        s = SimState(
            mode="static",
            arch="AMD64",
            dict_memory_backer=initial_memory,
            add_options={o.ABSTRACT_SOLVER, o.ABSTRACT_MEMORY},
        )
        se = s.solver

        def to_vs(region, offset):
            return s.solver.VS(s.arch.bits, region, 0, offset)

        # Load a single-byte constant from global region
        expr = s.memory.load(to_vs("global", 2), 1)
        assert s.solver.eval(expr) == 0x43
        assert s.solver.max_int(expr) == 0x43
        assert s.solver.min_int(expr) == 0x43

        # Store a single-byte constant to global region
        s.memory.store(to_vs("global", 1), s.solver.BVV(b"D"), 1)
        expr = s.memory.load(to_vs("global", 1), 1)
        assert s.solver.eval(expr) == 0x44

        # Store a single-byte StridedInterval to global region
        si_0 = s.solver.BVS("unnamed", 8, 10, 20, 2)
        s.memory.store(to_vs("global", 4), si_0)

        # Load the single-byte StridedInterval from global region
        expr = s.memory.load(to_vs("global", 4), 1)
        assert s.solver.min_int(expr) == 10
        assert s.solver.max_int(expr) == 20
        assert s.solver.eval_upto(expr, 100) == [10, 12, 14, 16, 18, 20]

        # Store a two-byte StridedInterval object to global region
        si_1 = s.solver.BVS("unnamed", 16, 10, 20, 2)
        s.memory.store(to_vs("global", 5), si_1)

        # Load the two-byte StridedInterval object from global region
        expr = s.memory.load(to_vs("global", 5), 2)
        assert claripy.backends.vsa.identical(expr, si_1)

        # Store a four-byte StridedInterval object to global region
        si_2 = s.solver.BVS("unnamed", 32, 8000, 9000, 2)
        s.memory.store(to_vs("global", 7), si_2)

        # Load the four-byte StridedInterval object from global region
        expr = s.memory.load(to_vs("global", 7), 4)
        assert claripy.backends.vsa.identical(expr, s.solver.BVS("unnamed", 32, 8000, 9000, 2))

        # Test default values
        s.options.remove(o.SYMBOLIC_INITIAL_VALUES)
        expr = s.memory.load(to_vs("global", 100), 4)
        assert claripy.backends.vsa.identical(expr, s.solver.BVS("unnamed", 32, 0, 0, 0))

        # Test default values (symbolic)
        s.options.add(o.SYMBOLIC_INITIAL_VALUES)
        expr = s.memory.load(to_vs("global", 104), 4)
        assert claripy.backends.vsa.identical(expr, s.solver.BVS("unnamed", 32, 0, 0xFFFFFFFF, 1))
        assert claripy.backends.vsa.identical(expr, s.solver.BVS("unnamed", 32, -0x80000000, 0x7FFFFFFF, 1))

        #
        # Merging
        #

        # Merging two one-byte values
        s.memory.store(to_vs("function_merge", 0), s.solver.BVS("unnamed", 8, 0x10, 0x10, 0))
        a = s.copy()
        a.memory.store(to_vs("function_merge", 0), s.solver.BVS("unnamed", 8, 0x20, 0x20, 0))

        b = s.merge(a)[0]
        expr = b.memory.load(to_vs("function_merge", 0), 1)
        assert claripy.backends.vsa.identical(expr, s.solver.BVS("unnamed", 8, 0x10, 0x20, 0x10))

        #  |  MO(value_0)  |
        #  |  MO(value_1)  |
        # 0x20          0x24
        # Merge one byte in value_0/1 means merging the entire MemoryObject
        a = s.copy()
        a.memory.store(
            to_vs("function_merge", 0x20), se.SI(bits=32, stride=0, lower_bound=0x100000, upper_bound=0x100000)
        )
        b = s.copy()
        b.memory.store(
            to_vs("function_merge", 0x20), se.SI(bits=32, stride=0, lower_bound=0x100001, upper_bound=0x100001)
        )
        c = a.merge(b)[0]
        expr = c.memory.load(to_vs("function_merge", 0x20), 4)
        assert claripy.backends.vsa.identical(
            expr, se.SI(bits=32, stride=1, lower_bound=0x100000, upper_bound=0x100001)
        )
        c_page = c.memory._regions["function_merge"]._pages[0]
        object_set = {
            c_page._get_object(0x20, 0),
            c_page._get_object(0x21, 0),
            c_page._get_object(0x22, 0),
            c_page._get_object(0x23, 0),
        }
        assert len(object_set) == 1

        a = s.copy()
        a.memory.store(
            to_vs("function_merge", 0x20), se.SI(bits=32, stride=0x100000, lower_bound=0x100000, upper_bound=0x200000)
        )
        b = s.copy()
        b.memory.store(
            to_vs("function_merge", 0x20), se.SI(bits=32, stride=0, lower_bound=0x300000, upper_bound=0x300000)
        )
        c = a.merge(b)[0]
        expr = c.memory.load(to_vs("function_merge", 0x20), 4)
        assert claripy.backends.vsa.identical(
            expr, se.SI(bits=32, stride=0x100000, lower_bound=0x100000, upper_bound=0x300000)
        )
        object_set = {
            c_page._get_object(0x20, 0),
            c_page._get_object(0x21, 0),
            c_page._get_object(0x22, 0),
            c_page._get_object(0x23, 0),
        }
        assert len(object_set) == 1

        #
        # Widening
        #

        a = s.solver.SI(bits=32, stride=1, lower_bound=1, upper_bound=2)
        b = s.solver.SI(bits=32, stride=1, lower_bound=1, upper_bound=3)
        a = a.reversed
        b = b.reversed
        # widened = a.widen(b)
        # TODO: Added a proper test case
        # print widened.reversed

        # We are done!
        # Restore the old claripy standalone object
        # claripy.set_claripy(old_claripy_standalone)

    def test_abstract_memory_find(self):
        initial_memory = {1: b"A", 2: b"B", 3: b"\x00"}

        s = SimState(
            mode="static",
            arch="AMD64",
            dict_memory_backer=initial_memory,
            add_options={o.ABSTRACT_SOLVER, o.ABSTRACT_MEMORY},
        )

        se = s.solver
        BVV = se.BVV
        VS = se.VS
        SI = se.SI

        s.memory.store(4, se.TSI(bits=64))

        def to_vs(region, offset):
            return VS(s.arch.bits, region, 0, offset)

        r, _, _ = s.memory.find(to_vs("global", 1), BVV(b"A"), 8)

        r_model = claripy.backends.vsa.convert(r)
        s_expected = claripy.backends.vsa.convert(SI(bits=64, to_conv=1))
        assert isinstance(r_model, claripy.vsa.ValueSet)
        assert list(r_model.regions.keys()) == ["global"]
        assert claripy.backends.vsa.identical(r_model.regions["global"], s_expected)

        r, _, _ = s.memory.find(to_vs("global", 1), BVV(b"B"), 8)
        r_model = claripy.backends.vsa.convert(r)
        s_expected = claripy.backends.vsa.convert(SI(bits=64, to_conv=2))
        assert isinstance(r_model, claripy.vsa.ValueSet)
        assert list(r_model.regions.keys()) == ["global"]
        assert claripy.backends.vsa.identical(r_model.regions["global"], s_expected)

        r, _, _ = s.memory.find(to_vs("global", 1), BVV(b"\0"), 8)
        r_model = claripy.backends.vsa.convert(r)
        s_expected = claripy.backends.vsa.convert(SI(bits=64, to_conv=3))
        assert isinstance(r_model, claripy.vsa.ValueSet)
        assert list(r_model.regions.keys()) == ["global"]
        assert claripy.backends.vsa.identical(r_model.regions["global"], s_expected)

        # Find in StridedIntervals
        r, _, _ = s.memory.find(to_vs("global", 4), BVV(b"\0"), 8)
        r_model = claripy.backends.vsa.convert(r)
        s_expected = claripy.backends.vsa.convert(SI(bits=64, stride=1, lower_bound=4, upper_bound=11))
        assert isinstance(r_model, claripy.vsa.ValueSet)
        assert list(r_model.regions.keys()) == ["global"]
        assert claripy.backends.vsa.identical(r_model.regions["global"], s_expected)

    def test_registers(self):
        s = SimState(arch="AMD64")
        expr = s.registers.load("rax")
        assert s.solver.symbolic(expr)

        s.registers.store("rax", 0x31)
        expr = s.registers.load("rax")
        assert not s.solver.symbolic(expr)
        assert s.solver.eval(expr) == 0x00000031

    def test_fullpage_write(self):
        s = SimState(arch="AMD64")
        a = s.solver.BVV(b"A" * 0x2000)
        s.memory.store(0, a)
        # assert len(s.memory.mem._pages) == 2
        # assert len(s.memory.mem._pages[0].keys()) == 0
        # assert len(s.memory.mem._pages[1].keys()) == 0
        assert s.memory.load(0, 0x2000) is a
        assert a.variables != s.memory.load(0x2000, 1).variables

        s = SimState(arch="AMD64")
        a = s.solver.BVV(b"A" * 2)
        s.memory.store(0x1000, a)
        s.memory.store(0x2000, a)
        assert a.variables == s.memory.load(0x2000, 1).variables
        assert a.variables == s.memory.load(0x2001, 1).variables
        assert a.variables != s.memory.load(0x2002, 1).variables

        s = SimState(arch="AMD64")
        x = s.solver.BVV(b"X")
        a = s.solver.BVV(b"A" * 0x1000)
        s.memory.store(1, x)
        s2 = s.copy()
        s2.memory.store(0, a)
        assert len(s.memory.changed_bytes(s2.memory)) == 0x1000

        s = SimState(arch="AMD64")
        s.memory._maximum_symbolic_size = 0x2000000
        a = s.solver.BVS("A", 0x1000000 * 8)
        s.memory.store(0, a)
        b = s.memory.load(0, 0x1000000)
        assert b is a

    def test_symbolic_write(self):
        s = SimState(arch="AMD64", add_options={o.SYMBOLIC_WRITE_ADDRESSES})
        x = s.solver.BVS("x", 64)
        y = s.solver.BVS("y", 64)
        a = s.solver.BVV(b"A" * 0x10)
        b = s.solver.BVV(b"B")
        c = s.solver.BVV(b"C")
        d = s.solver.BVV(b"D")

        s.memory.store(0x10, a)
        s.add_constraints(x >= 0x10, x < 0x20)
        s.memory.store(x, b)

        for i in range(0x10, 0x20):
            assert len(s.solver.eval_upto(s.memory.load(i, 1), 10)) == 2

        s.memory.store(x, c)
        for i in range(0x10, 0x20):
            assert len(s.solver.eval_upto(s.memory.load(i, 1), 10)) == 2

        s2 = s.copy()
        s2.add_constraints(y >= 0x10, y < 0x20)
        s2.memory.store(y, d)
        for i in range(0x10, 0x20):
            assert len(s2.solver.eval_upto(s2.memory.load(i, 1), 10)) == 3

    def test_concrete_memset(self):
        def _individual_test(state, base, val, size):
            # time it
            start = time.time()
            memset = SIM_PROCEDURES["libc"]["memset"]().execute(state, arguments=[base, state.solver.BVV(val, 8), size])
            elapsed = time.time() - start

            # should be done within 1 second
            assert elapsed <= 5
            # the result should be good
            byt_0 = memset.state.memory.load(base, 1)
            assert s.solver.eval_upto(byt_0, 10) == [val]
            byt_1 = memset.state.memory.load(base + 1, 1)
            assert s.solver.eval_upto(byt_1, 10) == [val]
            byt_2 = memset.state.memory.load(base + size - 1, 1)
            assert s.solver.eval_upto(byt_2, 10) == [val]

        BASE = 0x800000
        SIZE = 0x200000

        # Writes many zeros
        VAL = 0
        s = SimState(arch="AMD64")
        _individual_test(s, BASE, VAL, SIZE)

        # Writes many ones
        VAL = 1
        s = SimState(arch="AMD64")
        _individual_test(s, BASE, VAL, SIZE)

    def test_false_condition(self):
        s = SimState(arch="AMD64")

        asdf = s.solver.BVV(b"asdf")
        fdsa = s.solver.BVV(b"fdsa")
        s.memory.store(0x1000, asdf)
        s.memory.store(0x1000, fdsa, condition=s.solver.false)
        s.memory.store(0, fdsa, condition=s.solver.false)

        assert s.memory.load(0x1000, 4) is asdf
        assert 0 not in s.memory._pages

    def test_fast_memory(self):
        s = SimState(arch="AMD64", add_options={o.FAST_REGISTERS, o.FAST_MEMORY})

        s.regs.rax = 0x4142434445464748
        s.regs.rbx = 0x5555555544444444
        assert (s.regs.rax == 0x4142434445464748).is_true()
        assert (s.regs.rbx == 0x5555555544444444).is_true()

        self._concrete_memory_tests(s)

    def test_light_memory(self):
        s = SimState(arch="AMD64", plugins={"registers": SimLightRegisters()})
        assert type(s.registers) is SimLightRegisters

        assert s.regs.rax.symbolic
        s.regs.rax = 0x4142434445464748
        assert (s.regs.rax == 0x4142434445464748).is_true()

        assert s.regs.rbx.symbolic
        s.regs.rbx = 0x5555555544444444
        assert (s.regs.rbx == 0x5555555544444444).is_true()

        assert s.regs.rcx.symbolic

        s.regs.ah = 0
        assert (s.regs.rax == 0x4142434445460048).is_true()

        s.regs.cl = 0
        assert s.regs.rcx.symbolic

    def test_crosspage_store(self):
        for memcls in [
            UltraPageMemory,
            ListPageMemory,
        ]:
            state = SimState(arch="x86", mode="symbolic", plugins={"memory": memcls()})

            state.regs.sp = 0xBAAAFFFC
            state.memory.store(state.regs.sp, b"\x01\x02\x03\x04" + b"\x05\x06\x07\x08")
            assert state.solver.eval(state.memory.load(state.regs.sp, 8)) == 0x0102030405060708

            state.memory.store(state.regs.sp, b"\x01\x02\x03\x04" + b"\x05\x06\x07\x08", endness="Iend_LE")
            assert state.solver.eval(state.memory.load(state.regs.sp, 8)) == 0x0807060504030201

            symbol = claripy.BVS("symbol", 64)
            state.memory.store(state.regs.sp, symbol)
            assert state.memory.load(state.regs.sp, 8) is symbol

            state.memory.store(state.regs.sp, symbol, endness="Iend_LE")
            assert state.memory.load(state.regs.sp, 8) is symbol.reversed

    def test_mv_crosspage_store(self):
        for memcls in [
            MVPageMemory,
        ]:
            state = SimState(arch="x86", mode="symbolic", plugins={"memory": memcls()})

            mv = MultiValues(offset_to_values={0: {claripy.BVV(1337, 32)}, 4: {claripy.BVV(13371337, 8 * 5)}})
            state.memory.store(4096 - 3, mv)

            first_three_bytes = state.memory.load(4096 - 3, size=3)
            assert state.solver.eval_one(first_three_bytes.one_value()) == 1337 >> 8

            next_one_byte = state.memory.load(4096, size=1)
            assert state.solver.eval_one(next_one_byte.one_value()) == 1337 & 0xFF

            first_four_bytes = state.memory.load(4096 - 3, size=4)
            assert state.solver.eval_one(first_four_bytes.one_value()) == 1337

            all_bytes = state.memory.load(4096 - 3, size=9)
            assert state.solver.eval_one(all_bytes.one_value()) == (1337 << 40) | 13371337

            data = {
                0: {claripy.BVS("TOP", 32), claripy.BVV(0x56495254, 32)},
                4: {claripy.BVV(47, 8), claripy.BVV(85, 8)},
                5: {claripy.BVS("TOP", 32), claripy.BVV(0x414C5F53, 32)},
                9: {claripy.BVV(69, 8), claripy.BVV(47, 8)},
                10: {
                    claripy.BVV(0x52564552, 32),
                    claripy.BVV(0x56495254, 32),
                    claripy.BVS("TOP", 32),
                    claripy.BVV(0x63757272, 32),
                },
                14: {claripy.BVV(101, 8), claripy.BVV(47, 8), claripy.BVV(46, 8), claripy.BVV(85, 8)},
                15: {
                    claripy.BVV(0x656E74, 24),
                    claripy.BVV(0x6E7400, 24),
                    claripy.BVV(0x414C5F, 24),
                    claripy.BVS("TOP", 24),
                },
                18: {claripy.BVV(114, 8), claripy.BVS("TOP", 8), claripy.BVV(83, 8), claripy.BVV(47, 8)},
                19: {
                    claripy.BVV(0x45525645, 32),
                    claripy.BVS("TOP", 32),
                    claripy.BVV(0x2E747278, 32),
                    claripy.BVV(0x795B2564, 32),
                },
                23: {claripy.BVV(46, 8), claripy.BVV(82, 8), claripy.BVV(0, 8), claripy.BVV(93, 8)},
                24: {claripy.BVV(0x74727800, 32), claripy.BVV(0x2E69705B, 32), claripy.BVV(0x2E656E74, 32)},
                28: {
                    claripy.BVV(0x72795B25645D2E69705B25645D2E636F75, 136),
                    claripy.BVV(0x25645D2E636F756E74203D2025643B0A00, 136),
                },
                45: {claripy.BVV(110, 8), claripy.BVV(47, 8)},
                46: {claripy.BVV(0x74203D20, 32), claripy.BVS("TOP", 32)},
                50: {claripy.BVV(37, 8), claripy.BVV(47, 8)},
                51: {
                    claripy.BVS("TOP", 32),
                    claripy.BVV(0x56495254, 32),
                    claripy.BVV(0x63757272, 32),
                    claripy.BVV(0x643B0A00, 32),
                },
                55: {claripy.BVV(101, 8), claripy.BVV(85, 8), claripy.BVV(47, 8)},
                56: {claripy.BVV(0x6E7400, 24), claripy.BVV(0x414C5F, 24), claripy.BVS("TOP", 24)},
                59: {claripy.BVS("TOP", 8)},
                60: {claripy.BVV(0x45525645, 32), claripy.BVS("TOP", 32), claripy.BVV(0x2E747278, 32)},
                64: {claripy.BVV(46, 8), claripy.BVV(82, 8), claripy.BVV(0, 8)},
                65: {claripy.BVV(0x74727800, 32), claripy.BVV(0x2E656E74, 32)},
                69: {claripy.BVV(0x72795B25645D2E69705B25645D2E636F756E74203D2025643B0A00, 216)},
                96: {claripy.BVV(47, 8)},
                97: {claripy.BVS("TOP", 32)},
                101: {claripy.BVV(0x2E74727800, 40)},
            }
            mv = MultiValues(offset_to_values=data)
            state.memory.store(0x7FFEFF9C, mv)  # should not crash

    def test_crosspage_read(self):
        state = SimState(arch="ARM")
        state.regs.sp = 0x7FFF0008
        state.stack_push(0x44556677)
        state.stack_push(0x1)
        state.stack_push(0x2)
        state.stack_push(0x3)
        state.stack_push(0x4)
        state.stack_push(0x99887766)
        state.stack_push(0x5)
        state.stack_push(0x105C8)
        state.stack_push(0x11223344)

        r1 = state.memory.load(state.regs.sp, 36)
        assert bytes.fromhex("77665544") in state.solver.eval(r1, cast_to=bytes)

        state.stack_push(0x10564)

        r2 = state.memory.load(state.regs.sp, 40)
        assert bytes.fromhex("77665544") in state.solver.eval(r2, cast_to=bytes)
        # assert s.solver.eval(r, 2) == ( 0xffeeddccbbaa998877665544, )

    def test_address_wrap(self):
        for memcls in [UltraPageMemory, ListPageMemory]:
            state = SimState(arch="x86", mode="symbolic", plugins={"memory": memcls()})
            symbol = claripy.BVS("symbol", 64)

            state.memory.store(0xFFFFFFFF, symbol.get_byte(0))
            state.memory.store(0, b"\0" * 8)
            assert len(state.memory.load(0xFFFFFFFF, 8)) == 64

            state.memory.store(0xFFFFFFFF, b"ABCD")
            assert state.solver.eval(state.memory.load(0, 3)) == 0x424344

            state.memory.store(0xFFFFFFFF, symbol)
            assert state.memory.load(0, 1) is symbol[64 - 8 - 1 : 64 - 16]

    def test_underconstrained(self):
        state = SimState(arch="AMD64", add_options={o.UNDER_CONSTRAINED_SYMEXEC})

        # test that under-constrained load is constrained
        ptr1 = state.memory.load(0x4141414141414000, size=8, endness="Iend_LE")
        assert ptr1.uc_alloc_depth == 0
        assert ptr1.uninitialized
        state.memory.load(ptr1, size=1)
        # ptr1 should have been constrained
        assert state.solver.min_int(ptr1) == state.solver.max_int(ptr1)

        # test that under-constrained store is constrained
        ptr2 = state.memory.load(0x4141414141414008, size=8, endness="Iend_LE")
        assert ptr2.uc_alloc_depth == 0
        assert ptr2.uninitialized
        state.memory.store(ptr2, b"\x41", size=1)
        # ptr2 should have been constrained
        assert state.solver.min_int(ptr2) == state.solver.max_int(ptr2)

        # ptr1 and ptr2 should not point to the same region
        assert state.solver.eval(ptr1) != state.solver.eval(ptr2)

        # uninitialized load and stores w/o uc_alloc_depth should not crash
        ptr3 = claripy.Concat(
            state.memory.load(0x4141414141414010, size=4, endness="Iend_LE"),
            state.memory.load(0x4141414141414014, size=4, endness="Iend_LE"),
        )
        assert ptr3.uninitialized
        assert ptr3.uc_alloc_depth is None  # because uc_alloc_depth doesn't carry across Concat
        # we don't care what these do, as long as they don't crash
        state.memory.store(ptr3, b"\x41", size=1)
        state.memory.load(ptr3, size=1)

    def test_concrete_load_non_adjacent_pages(self):
        s = SimState(arch="AMD64", mode="symbolic", plugins={"memory": UltraPageMemory()})

        s.memory.store(0x100000, b"\x01" * 4096)
        s.memory.store(0x100000 + 4096, b"\x02" * 4096)
        mv = s.memory.concrete_load(0x100000 + 0xFFA, 400)
        assert len(mv) == 400, "Loading data across non-physically adjacent pages failed for ultra pages."
        assert mv == (b"\x01" * 6) + (b"\x02" * 394)

    def test_hex_dump(self):
        s = SimState(arch="AMD64")
        addr = s.heap.allocate(0x20)
        s.memory.store(addr, claripy.Concat(claripy.BVV("ABCDEFGH"), claripy.BVS("symbolic_part", 24 * s.arch.bits)))
        dump = s.memory.hex_dump(addr, 0x20)
        assert (
            dump == "c0000000: 41424344 45464748 ???????? ???????? ABCDEFGH????????\n"
            "c0000010: ???????? ???????? ???????? ???????? ????????????????\n"
        )

        dump = s.memory.hex_dump(
            addr,
            0x20,
            extra_constraints=(s.memory.load(addr + 0x10, 4) == 0xDEADBEEF,),
            solve=True,
            endianness="Iend_LE",
        )
        assert (
            dump == "c0000000: 44434241 48474645 ???????? ???????? ABCDEFGH????????\n"
            "c0000010: efbeadde ???????? ???????? ???????? ....????????????\n"
        )

    def test_concrete_load(self):
        for memcls in [UltraPageMemory, ListPageMemory]:
            state = SimState(arch="AMD64", mode="symbolic", plugins={"memory": memcls()})
            state.memory.store(0x20000, b"aaaabbbbccccdddd")

            data, bitmap = state.memory.concrete_load(0x20000, 4, with_bitmap=True)
            data_bytes = bytes(d if b == 0 else 0 for d, b in zip(data, bitmap))
            assert data_bytes == b"aaaa"
            assert bitmap.tobytes() == b"\x00\x00\x00\x00"

            data, bitmap = state.memory.concrete_load(0x20004, 8, with_bitmap=True)
            data_bytes = bytes(d if b == 0 else 0 for d, b in zip(data, bitmap))
            assert data_bytes == b"bbbbcccc"
            assert bitmap.tobytes() == b"\x00\x00\x00\x00\x00\x00\x00\x00"

            state.memory.store(0x20001, claripy.BVS("flag", 8))
            data, bitmap = state.memory.concrete_load(0x20000, 4, with_bitmap=True)
            data_bytes = bytes(d if b == 0 else 0 for d, b in zip(data, bitmap))
            assert data_bytes == b"a\x00aa"
            assert bitmap.tobytes() == b"\x00\x01\x00\x00"

            expr = claripy.Concat(
                claripy.BVS("flag_0", 1),
                claripy.BVS("flag_1", 2),
                claripy.BVS("flag_2", 3),
                claripy.BVS("flag_3", 6),
                claripy.BVS("flag_4", 4),
            )
            state.memory.store(0x20001, expr)
            data, bitmap = state.memory.concrete_load(0x20000, 4, with_bitmap=True)
            data_bytes = bytes(d if b == 0 else 0 for d, b in zip(data, bitmap))
            assert data_bytes == b"a\x00\x00a"
            assert bitmap.tobytes() == b"\x00\x01\x01\x00"

            expr = claripy.Concat(
                claripy.BVS("flag_0", 1),
                claripy.BVV(1, 2),
                claripy.BVV(3, 3),
                claripy.BVV(6, 6),
                claripy.BVS("flag_4", 4),
            )
            state.memory.store(0x20005, expr)
            data, bitmap = state.memory.concrete_load(0x20004, 4, with_bitmap=True)
            data_bytes = bytes(d if b == 0 else 0 for d, b in zip(data, bitmap))
            assert data_bytes == b"b\x00\x00b"
            assert bitmap.tobytes() == b"\x00\x01\x01\x00"

            expr = claripy.Concat(
                claripy.BVV(7, 7),
                claripy.BVS("flag_0", 2),
                claripy.BVV(7, 7),
            )
            state.memory.store(0x20005, expr)
            data, bitmap = state.memory.concrete_load(0x20004, 4, with_bitmap=True)
            data_bytes = bytes(d if b == 0 else 0 for d, b in zip(data, bitmap))
            assert data_bytes == b"b\x00\x00b"
            assert bitmap.tobytes() == b"\x00\x01\x01\x00"

            expr = claripy.Concat(
                claripy.BVV(1, 1),
                claripy.BVS("flag_0", 14),
                claripy.BVV(1, 1),
            )
            state.memory.store(0x20005, expr)
            data, bitmap = state.memory.concrete_load(0x20004, 4, with_bitmap=True)
            data_bytes = bytes(d if b == 0 else 0 for d, b in zip(data, bitmap))
            assert data_bytes == b"b\x00\x00b"
            assert bitmap.tobytes() == b"\x00\x01\x01\x00"

    def test_multivalued_list_page(self):
        state = SimState(arch="AMD64", mode="symbolic", plugins={"memory": MultiValuedMemory()})

        # strong update
        state.memory.store(0x100, claripy.BVV(0x40, 64))
        state.memory.store(0x100, claripy.BVV(0x80818283, 64))
        a = state.memory.load(0x100, size=8).one_value()
        assert a is not None
        assert a is claripy.BVV(0x80818283, 64)

        # strong update with partial overwrites
        state.memory.store(0x100, claripy.BVV(0x0, 64))
        state.memory.store(0x104, claripy.BVV(0x1337, 32))
        a = state.memory.load(0x100, size=8).one_value()
        assert a is not None
        assert a is claripy.BVV(0x1337, 64)

        # weak updates
        state.memory.store(0x120, claripy.BVV(0x40, 64))
        state.memory.store(0x120, claripy.BVV(0x85868788, 64), weak=True)
        a = state.memory.load(0x120, size=8)
        assert len(a.keys()) == 1
        assert 0 in a
        assert len(list(a.values())[0]) == 2
        assert {state.solver.eval_exact(item, 1)[0] for item in list(a.values())[0]} == {0x40, 0x85868788}

        # weak updates with symbolic values
        A = claripy.BVS("a", 64)
        state.memory.store(0x140, A, endness=state.arch.memory_endness)
        state.memory.store(0x141, claripy.BVS("b", 64), endness=state.arch.memory_endness, weak=True)
        a = state.memory.load(0x140, size=8)
        assert len(a.keys()) == 2
        assert 0 in a
        assert 1 in a
        assert len(list(a.values())[0]) == 1
        assert next(iter(a[0])) is A[7:0]
        assert len(a[1]) == 2

    def test_conditional_concretization(self):
        class ZeroFillerMemory(UltraPageMemory):
            def _default_value(self, *args, size=None, **kwargs):  # pylint:disable=unused-argument
                return self.state.solver.BVV(0, size)

        state = SimState(arch="AMD64", mode="symbolic", plugins={"memory": ZeroFillerMemory()})
        state.options.add(o.ZERO_FILL_UNCONSTRAINED_MEMORY)

        cond = state.solver.BoolS("cond")
        addr = state.solver.BVS("addr", 32)

        state.memory.store(addr, 0x12345678, size=4, condition=cond, endness=state.arch.memory_endness)
        val = state.memory.load(addr, size=4, condition=cond, endness=state.arch.memory_endness)
        assert set(state.solver.eval_upto(cond, 2)) == {True, False}
        assert (val == 0x12345678).is_true()


if __name__ == "__main__":
    unittest.main()



from typing import List, Tuple, Set, Dict, Union, Optional, Any

import claripy

from angr.storage.memory_object import SimMemoryObject, SimLabeledMemoryObject
from .multi_values import MultiValues


class CooperationBase:
    """
    Any given subclass of this class which is not a subclass of MemoryMixin should have the property that any subclass
    it which *is* a subclass of MemoryMixin should all work with the same datatypes
    """

    @classmethod
    def _compose_objects(cls, objects, size, endness, **kwargs):
        """
        Provide this a list of the result of several load calls, and it will compose them into a single result.
        """

    @classmethod
    def _decompose_objects(cls, addr, data, endness, **kwargs) -> Tuple[Any, int, int]:
        """
        A bidirectional generator. No idea if this is overengineered. Usage is that you send it a size to use
        and it yields a tuple of three elements: the object to store for the next n bytes, the base address of the
        object, and the size of the object.
        """

    @classmethod
    def _zero_objects(cls, addr, size, **kwargs):
        """
        Like decompose objects, but with a size to zero-fill instead of explicit data
        """

    @classmethod
    def _force_store_cooperation(cls, addr, data, size, endness, **kwargs):
        if data is not None:
            sub_gen = cls._decompose_objects(addr, data, endness, **kwargs)
        else:
            sub_gen = cls._zero_objects(addr, size, **kwargs)

        next(sub_gen)
        sub_data, _, _ = sub_gen.send(size)
        sub_gen.close()
        return sub_data

    @classmethod
    def _force_load_cooperation(cls, results, size, endness, **kwargs):
        return cls._compose_objects([results], size, endness, **kwargs)


class MemoryObjectMixin(CooperationBase):
    """
    Uses SimMemoryObjects in region storage.
    With this, load will return a list of tuple (address, MO) and store will take a MO.
    """

    @classmethod
    def _compose_objects(
        cls,
        objects: List[List[Tuple[int, SimMemoryObject]]],
        size,
        endness=None,
        memory=None,
        labels: Optional[List] = None,
        **kwargs,
    ):
        c_objects = []
        for objlist in objects:
            for element in objlist:
                if not c_objects or element[1] is not c_objects[-1][1]:
                    c_objects.append(element)

        mask = (1 << memory.state.arch.bits) - 1
        if labels is None:
            # fast path - ignore labels
            elements = [
                o.bytes_at(
                    a,
                    ((c_objects[i + 1][0] - a) & mask)
                    if i != len(c_objects) - 1
                    else ((c_objects[0][0] + size - a) & mask),
                    endness=endness,
                )
                for i, (a, o) in enumerate(c_objects)
            ]
        else:
            # we need to extract labels for SimLabeledMemoryObjects
            elements = []
            offset = 0
            for i, (a, o) in enumerate(c_objects):
                length: int = (
                    ((c_objects[i + 1][0] - a) & mask)
                    if i != len(c_objects) - 1
                    else ((c_objects[0][0] + size - a) & mask)
                )
                byts = o.bytes_at(a, length, endness=endness)
                elements.append(byts)
                if isinstance(o, SimLabeledMemoryObject):
                    labels.append((offset, a - o.base, length, o.label))
                offset += length
        if len(elements) == 0:
            # nothing is read out
            return claripy.BVV(0, 0)
        if len(elements) == 1:
            return elements[0]

        if endness == "Iend_LE":
            elements = list(reversed(elements))

        return elements[0].concat(*elements[1:])

    @classmethod
    def _decompose_objects(cls, addr, data, endness, memory=None, page_addr=0, label=None, **kwargs):
        # the generator model is definitely overengineered here but wouldn't be if we were working with raw BVs
        cur_addr = addr + page_addr
        byte_width = memory.state.arch.byte_width if memory is not None else 8
        if label is None:
            memory_object = SimMemoryObject(data, cur_addr, endness, byte_width=byte_width)
        else:
            memory_object = SimLabeledMemoryObject(data, cur_addr, endness, byte_width=byte_width, label=label)

        if data.symbolic and data.op == "Concat":
            next_elem_size_left = data.args[0].size() // 8
            next_elem_index = 0

        size = yield
        max_size = kwargs.get("max_size", size)
        while True:
            if data.symbolic and data.op == "Concat" and data.size() > max_size:
                # Generate new memory object with only size bytes to speed up extracting bytes
                cur_data_size_bits = 0
                requested_size_bits = size * 8
                cur_data = []
                while cur_data_size_bits < requested_size_bits:
                    if next_elem_size_left == 0:
                        next_elem_index += 1

                    next_elem = data.args[next_elem_index]
                    cur_data.append(next_elem)
                    next_elem_size_left = next_elem.size()
                    added_size = min(requested_size_bits - cur_data_size_bits, next_elem.size())
                    cur_data_size_bits += added_size
                    next_elem_size_left = next_elem_size_left - added_size

                cur_data = claripy.Concat(*cur_data)
                if label is None:
                    memory_object = SimMemoryObject(cur_data, cur_addr, endness, byte_width=byte_width)
                else:
                    memory_object = SimLabeledMemoryObject(
                        cur_data, cur_addr, endness, byte_width=byte_width, label=label
                    )
            cur_addr += size
            size = yield memory_object, memory_object.base, memory_object.length

    @classmethod
    def _zero_objects(cls, addr, size, memory=None, **kwargs):
        data = claripy.BVV(0, size * memory.state.arch.byte_width if memory is not None else 8)
        return cls._decompose_objects(addr, data, "Iend_BE", memory=memory, **kwargs)


class MemoryObjectSetMixin(CooperationBase):
    """
    Uses sets of SimMemoryObjects in region storage.
    """

    @classmethod
    def _compose_objects(
        cls, objects: List[List[Tuple[int, Set[SimMemoryObject]]]], size, endness=None, memory=None, **kwargs
    ):
        c_objects: List[Tuple[int, Union[SimMemoryObject, Set[SimMemoryObject]]]] = []
        for objlist in objects:
            for element in objlist:
                if not c_objects or element[1] is not c_objects[-1][1]:
                    c_objects.append(element)

        mask = (1 << memory.state.arch.bits) - 1
        elements: List[Set[claripy.ast.Base]] = []
        for i, (a, objs) in enumerate(c_objects):
            chopped_set = set()
            if type(objs) is not set:
                objs = {objs}
            for o in objs:
                if o.includes(a):
                    chopped = o.bytes_at(
                        a,
                        ((c_objects[i + 1][0] - a) & mask)
                        if i != len(c_objects) - 1
                        else ((c_objects[0][0] + size - a) & mask),
                        endness=endness,
                    )
                    chopped_set.add(chopped)
            if chopped_set:
                elements.append(chopped_set)

        if len(elements) == 0:
            # nothing is read out
            return MultiValues(claripy.BVV(0, 0))

        if len(elements) == 1:
            if len(elements[0]) == 1:
                return MultiValues(next(iter(elements[0])))
            return MultiValues(offset_to_values={0: elements[0]})

        if endness == "Iend_LE":
            elements = list(reversed(elements))

        mv = MultiValues()
        offset = 0
        start_offset = 0
        prev_value = ...
        for i, value_set in enumerate(elements):
            if len(value_set) == 1:
                if prev_value is ...:
                    prev_value = next(iter(value_set))
                    start_offset = offset
                else:
                    prev_value = prev_value.concat(next(iter(value_set)))
            else:
                if prev_value is not ...:
                    mv.add_value(start_offset, prev_value)
                    prev_value = ...

                for value in value_set:
                    mv.add_value(offset, value)

            offset += next(iter(value_set)).size() // memory.state.arch.byte_width

        if prev_value is not ...:
            mv.add_value(start_offset, prev_value)
            prev_value = ...

        return mv

    @classmethod
    def _decompose_objects(cls, addr, data, endness, memory=None, page_addr=0, label=None, **kwargs):
        # the generator model is definitely overengineered here but wouldn't be if we were working with raw BVs
        cur_addr = addr + page_addr
        if isinstance(data, MultiValues):
            # for MultiValues, we return sets of SimMemoryObjects
            assert label is None  # TODO: Support labels

            size = yield
            offset_to_mos: Dict[int, Set[SimMemoryObject]] = {}
            for offset, vs in data.items():
                offset_to_mos[offset] = set()
                for v in vs:
                    offset_to_mos[offset].add(
                        SimMemoryObject(
                            v,
                            cur_addr + offset,
                            endness,
                            byte_width=memory.state.arch.byte_width if memory is not None else 0,
                        )
                    )

            sorted_offsets = list(sorted(offset_to_mos.keys()))
            pos = 0
            while pos < len(sorted_offsets):
                mos = set(offset_to_mos[sorted_offsets[pos]])
                first_mo = next(iter(mos))
                old_size = size

                size = yield mos, first_mo.base, first_mo.length
                cur_addr += min(first_mo.length, old_size)
                if sorted_offsets[pos] + first_mo.length <= cur_addr - addr - page_addr:
                    pos += 1

        else:
            if label is None:
                obj = SimMemoryObject(
                    data, cur_addr, endness, byte_width=memory.state.arch.byte_width if memory is not None else 8
                )
            else:
                obj = SimLabeledMemoryObject(
                    data,
                    cur_addr,
                    endness,
                    byte_width=memory.state.arch.byte_width if memory is not None else 8,
                    label=label,
                )
            _ = yield
            while True:
                _ = yield {obj}, obj.base, obj.length

    @classmethod
    def _zero_objects(cls, addr, size, memory=None, **kwargs):
        data = claripy.BVV(0, size * memory.state.arch.byte_width if memory is not None else 8)
        return cls._decompose_objects(addr, data, "Iend_BE", memory=memory, **kwargs)


class BasicClaripyCooperation(CooperationBase):
    """
    Mix this (along with PageBase) into a storage class which supports loading and storing claripy bitvectors and it
    will be able to work as a page in the paged memory model.
    """

    @classmethod
    def _compose_objects(cls, objects, size, endness, **kwargs):
        if endness == "Iend_LE":
            objects = reversed(objects)

        return claripy.Concat(*objects)

    @classmethod
    def _decompose_objects(cls, addr, data, endness, memory=None, **kwargs):
        if endness == "Iend_BE":
            size = yield
            offset = 0
            while True:
                data_slice = data.get_bytes(offset, size)
                data_slide_base = addr + offset
                offset += size
                size = yield data_slice, data_slide_base, data_slice.length
        else:
            size = yield
            offset = len(data) // (memory.state.arch.byte_width if memory is not None else 8)
            while True:
                offset -= size
                data_slice = data.get_bytes(offset, size)
                size = yield data_slice, addr + offset, data_slice.length

    @classmethod
    def _zero_objects(cls, addr, size, memory=None, **kwargs):
        data = claripy.BVV(0, size * memory.state.arch.byte_width if memory is not None else 8)
        return cls._decompose_objects(addr, data, "Iend_BE", memory=memory, **kwargs)



#!/usr/bin/env python3
# pylint:disable=missing-class-docstring,no-self-use
from unittest import TestCase, main

import angr
from angr.analyses import Disassembly
from angr.analyses.disassembly import MemoryOperand, Instruction


class TestDisassembly(TestCase):
    def test_arm64_dissect_instructions(self):
        proj = angr.load_shellcode(
            b"\x00\xe4\x00\x6f"
            b"\x43\x3c\x0b\x0e"
            b"\x54\x9a\xb7\x72"
            b"\xfc\x6f\xba\xa9"
            b"\x88\x03\x98\x1a"
            b"\x00\x60\x01\x4e",
            "AARCH64",
            0,
        )
        # movi   v0.2d, #0000000000000000'                              ; SIMD register
        # umov   w3, v2.b[5]                                            ; SIMD register index
        # movk   w20, #0xbcd2, lsl #16                                  ; ARM64 shifter
        # stp    x28, x27, [sp, #-0x60]!                                ; ARM64 pre-indexed operand
        # csel   w8, w28, w24, eq                                       ; Condition code at the end
        # tbl    v0.16b, {v0.16b, v1.16b, v2.16b, v3.16b}, v1.16b       ; Multiple SIMD regs in table
        block = proj.factory.block(0)
        disasm = proj.analyses[Disassembly].prep()(ranges=[(block.addr, block.addr + block.size)])

        insns = [r for r in disasm.raw_result if isinstance(r, Instruction)]
        rendered_insns = [i.render()[0].lower() for i in insns]
        assert "v0.2d" in rendered_insns[0]
        assert "v2.b[5]" in rendered_insns[1]
        assert "lsl#16" in rendered_insns[2].replace(" ", "")
        assert rendered_insns[3].endswith("]!")
        assert rendered_insns[4].endswith("eq")
        insn = rendered_insns[5]
        regs_table = insn[insn.index("{") + 1 : insn.index("}")].replace(" ", "").split(",")
        assert ["v0.16b", "v1.16b", "v2.16b", "v3.16b"] == regs_table

    def test_arm32_dissect_instructions(self):
        proj = angr.load_shellcode(
            b"\x00\xc0\x2d\xe9\x10\xf9\xf9\xe9",
            "ARM",
            0,
        )
        # push    {lr, pc}
        # ldmib   sb!, {r4, r8, fp, ip, sp, lr, pc}^

        block = proj.factory.block(0)
        disasm = proj.analyses[Disassembly].prep()(ranges=[(block.addr, block.addr + block.size)])
        insns = [r for r in disasm.raw_result if isinstance(r, Instruction)]
        rendered_insns = [i.render()[0].lower() for i in insns]
        assert all(i in rendered_insns[0] for i in ("{", "}", "lr", "pc"))
        assert "sb!" in rendered_insns[1]
        assert rendered_insns[1].endswith("^")

    def test_arm32_thumb_dissect_instructions(self):
        proj = angr.load_shellcode(b"\x00\xf9\x01\x1a", "ARM", 0, thumb=True)
        # vst1.8  {d1, d2}, [r0], r1

        block = proj.factory.block(0, thumb=True)
        disasm = proj.analyses[Disassembly].prep()(ranges=[(block.addr, block.addr + block.size)], thumb=True)
        insns = [r for r in disasm.raw_result if isinstance(r, Instruction)]

        disassembly_operands = insns[0].operands
        capstone_operands = insns[0].insn.operands
        assert len(disassembly_operands) == len(capstone_operands)

    def test_mips32_missing_offset_in_instructions(self):
        proj = angr.load_shellcode(
            b"\x8f\xbc\x00\x10"
            b"\x02\x20\x30\x21"
            b"\x8F\x85\x80\x28"
            b"\x8F\x99\x81\x20"
            b"\x02\x40\x38\x21"
            b"\x24\xA5\x5E\x38"
            b"\x03\x20\xF8\x09"
            b"\x24\x04\x00\x02",
            "MIPS32",
            0,
        )
        # 0x0:    lw      $gp, 0x10($sp)
        # 0x4:    move    $a2, $s1
        # 0x8:    lw      $a1, -0x7fd8($gp)
        # 0xc:    lw      $t9, -0x7ee0($gp)
        # 0x10:   move    $a3, $s2
        # 0x14:   addiu   $a1, $a1, 0x5e38
        # 0x18:   jalr    $t9
        # 0x1c:   addiu   $a0, $zero, 2

        block = proj.factory.block(0)
        disass = proj.analyses[Disassembly].prep()(ranges=[(block.addr, block.addr + block.size)])
        result = disass.raw_result
        assert len(result) == 10, f"Incorrect number of instructions ({len(result)})"

        ins = result[4]
        operand_1 = ins.operands[1]
        assert isinstance(operand_1, MemoryOperand)
        assert len(operand_1.children) == 4
        assert len(operand_1.values) == 1
        assert len(operand_1.offset) == 1
        assert operand_1.offset_location == "prefix"
        rendered = disass.render(color=False)
        assert (
            rendered
            == """   _start:
0  lw      $gp, 0x10($sp)
4  move    $a2, $s1
8  lw      $a1, -0x7fd8($gp)
c  lw      $t9, -0x7ee0($gp)
10  move    $a3, $s2
14  addiu   $a1, $a1, 0x5e38
18  jalr    $t9
1c  addiu   $a0, $zero, 0x2"""
        )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# pylint: disable=missing-class-docstring,disable=no-self-use
__package__ = __package__ or "tests.state_plugins.posix"  # pylint:disable=redefined-builtin

import os
import unittest

import angr
from angr.state_plugins.posix import Flags

from ...common import bin_location


test_location = os.path.join(bin_location, "tests")


class TestFile(unittest.TestCase):
    def test_files(self):
        s = angr.SimState(arch="AMD64")
        s.posix.get_fd(1).write_data(b"HELLO")
        s.posix.get_fd(1).write_data(b"WORLD")
        assert s.posix.dumps(1) == b"HELLOWORLD"
        assert s.posix.stdout.concretize() == [b"HELLO", b"WORLD"]

        s = angr.SimState(arch="AMD64")
        s.posix.get_fd(1).write_data(b"A" * 0x1000, 0x800)
        assert s.posix.dumps(1) == b"A" * 0x800

    def test_file_read_missing_content(self):
        # test in tracing mode since the Reverse operator will not be optimized away
        s = angr.SimState(arch="AMD64", mode="tracing")
        fd = s.posix.open(b"/tmp/oops", Flags.O_RDWR)
        length = s.posix.get_fd(fd).read(0xC00000, 100)

        data = s.memory.load(0xC00000, length, endness="Iend_BE")
        assert data.op != "Reverse", "Byte strings read directly out of a file should not have Reverse operators."
        assert data.op == "BVS"
        assert len(data.variables) == 1
        assert "oops" in next(iter(data.variables))

    def test_concrete_fs_resolution(self):
        bin_path = os.path.join(test_location, "i386", "fauxware")
        proj = angr.Project(bin_path, auto_load_libs=False)
        state = proj.factory.entry_state(concrete_fs=True)
        fd = state.posix.open(bin_path, Flags.O_RDONLY)
        stat = state.posix.fstat(fd)
        size = stat.st_size
        int_size = state.solver.eval(size)

        assert stat
        assert int_size != 0
        assert not state.solver.symbolic(size)

    def test_sim_fs_resolution(self):
        bin_path = os.path.join(test_location, "i386", "fauxware")
        proj = angr.Project(bin_path, auto_load_libs=False)
        state = proj.factory.entry_state()
        fd = state.posix.open(bin_path, Flags.O_RDONLY)
        stat = state.posix.fstat(fd)
        size = stat.st_size

        assert stat
        assert state.solver.symbolic(size)


if __name__ == "__main__":
    unittest.main()



import angr


class KiUserExceptionDispatcher(angr.SimProcedure):
    local_vars = ("tib_ptr", "top_record", "cur_ptr")
    IS_FUNCTION = False

    tib_ptr = None
    top_record = None
    cur_ptr = None

    def run(self, record, context):
        if self.state.arch.name != "X86":
            raise angr.errors.SimUnsupportedError("KiUserDispatchException is only implemented for X86")

        self.tib_ptr = self.state.regs._fs.concat(self.state.solver.BVV(0, 16))
        self.top_record = self.state.mem[self.tib_ptr].uint32_t.resolved
        self.cur_ptr = self.top_record

        # set magic value to detect nested exceptions
        self.state.mem[self.tib_ptr].long = 0xBADFACE

        self.dispatch(record, context)

    def dispatch(self, record, context):
        if self.call_ret_expr is not None:
            try:
                disposition = self.state.solver.eval_one(self.call_ret_expr)
            except angr.errors.SimSolverError:
                raise angr.errors.SimError("Exception handler returned symbolic value %s" % self.call_ret_expr)
            if disposition == 0:  # Handled!!!
                self.project.simos._load_regs(self.state, context)
                # TODO: re-set the exception handler somehow?
                # self.state.mem[self.tib_ptr].uint32_t
                self.jump(self.state.regs._ip)
                return
            elif disposition == 1:  # unhandled, continue search
                pass
            elif disposition == 2:
                raise angr.errors.SimUnsupportedError("Exception disposition ExceptionNestedException is unsupported")
            elif disposition == 3:
                raise angr.errors.SimUnsupportedError("Exception disposition ExceptionCollidedUnwind is unsupported")
            else:
                raise angr.errors.SimError("Bad exception disposition %d" % disposition)

        # todo: check cur_ptr against stack bounds
        cur_ptr = self.cur_ptr
        if self.state.solver.is_true(cur_ptr == -1):
            raise angr.errors.SimError("Unhandled exception - exhausted all exception handlers")
        next_ptr = self.state.mem[self.cur_ptr].uint32_t.resolved
        func_ptr = self.state.mem[self.cur_ptr + 4].uint32_t.resolved

        self.cur_ptr = next_ptr
        # as far as I can tell it doesn't actually matter whether the callback is stdcall or cdecl
        self.call(func_ptr, (record, cur_ptr, context, 0xBADF00D), "dispatch", prototype="void x(int, int, int, int)")
        # bonus! after we've done the call, mutate the state even harder so ebp is pointing to some fake args
        self.successors.successors[0].regs.ebp = self.successors.successors[0].regs.esp - 4



#!/usr/bin/env python3
# pylint: disable=missing-class-docstring,no-self-use,line-too-long
__package__ = __package__ or "tests.analyses"  # pylint:disable=redefined-builtin

import os
import unittest

import angr
from angr.knowledge_plugins.xrefs import XRef, XRefType

from ..common import bin_location


test_location = os.path.join(bin_location, "tests")


class TestXrefs(unittest.TestCase):
    def test_lwip_udpecho_bm(self):
        bin_path = os.path.join(test_location, "armel", "lwip_udpecho_bm.elf")
        p = angr.Project(bin_path, auto_load_libs=False)
        cfg = p.analyses.CFG(data_references=True)

        func = cfg.functions[0x23C9]
        state = p.factory.blank_state()
        prop = p.analyses.Propagator(func=func, base_state=state)
        _ = p.analyses.XRefs(func=func, replacements=prop.replacements)

        timenow_cp_xrefs = p.kb.xrefs.get_xrefs_by_dst(0x23D4)  # the constant in the constant pool
        timenow_xrefs = p.kb.xrefs.get_xrefs_by_dst(0x1FFF36F4)  # the value in .bss

        assert len(timenow_cp_xrefs) == 1
        assert next(iter(timenow_cp_xrefs)) == XRef(ins_addr=0x23C9, dst=0x23D4, xref_type=XRefType.Read)

        assert len(timenow_xrefs) == 3
        assert [x for x in timenow_xrefs if x.type == XRefType.Offset][0] == XRef(
            ins_addr=0x23C9, dst=0x1FFF36F4, xref_type=XRefType.Offset
        )
        assert [x for x in timenow_xrefs if x.type == XRefType.Read][0] == XRef(
            ins_addr=0x23CB, dst=0x1FFF36F4, xref_type=XRefType.Read
        )
        assert [x for x in timenow_xrefs if x.type == XRefType.Write][0] == XRef(
            ins_addr=0x23CF, dst=0x1FFF36F4, xref_type=XRefType.Write
        )

    def test_lwip_udpecho_bm_the_better_way(self):
        bin_path = os.path.join(test_location, "armel", "lwip_udpecho_bm.elf")
        p = angr.Project(bin_path, auto_load_libs=False)
        p.analyses.CFG(cross_references=True)

        timenow_cp_xrefs = p.kb.xrefs.get_xrefs_by_dst(0x23D4)  # the constant in the constant pool
        timenow_xrefs = p.kb.xrefs.get_xrefs_by_dst(0x1FFF36F4)  # the value in .bss

        assert len(timenow_cp_xrefs) == 1
        assert next(iter(timenow_cp_xrefs)) == XRef(ins_addr=0x23C9, dst=0x23D4, xref_type=XRefType.Read)
        # sys_now (2), time_isr (3) == 5
        assert len(timenow_xrefs) == 5

    def test_p2im_drone_with_inits(self):
        bin_path = os.path.join(test_location, "armel", "p2im_drone.elf")
        proj = angr.Project(bin_path, auto_load_libs=False)
        cfg = proj.analyses.CFG(data_references=True)

        func = cfg.functions["Peripherals_Init"]
        state = proj.factory.blank_state()
        prop = proj.analyses.Propagator(func=func, base_state=state)

        init_finder = proj.analyses.InitializationFinder(func=func, replacements=prop.replacements)
        overlay_state = init_finder.overlay_state

        cfg.do_full_xrefs(overlay_state=overlay_state)

        h12c1_inst_xrefs = proj.kb.xrefs.get_xrefs_by_dst(0x20001500)
        assert len(h12c1_inst_xrefs) == 5


if __name__ == "__main__":
    unittest.main()



from typing import Optional, TYPE_CHECKING
import logging

from ...engines.light import SimEngineLight
from ...errors import SimEngineError

if TYPE_CHECKING:
    from angr.analyses.reaching_definitions.reaching_definitions import ReachingDefinitionsModel

l = logging.getLogger(name=__name__)


class SimEnginePropagatorBase(SimEngineLight):  # pylint:disable=abstract-method
    def __init__(
        self,
        stack_pointer_tracker=None,
        project=None,
        propagate_tmps=True,
        arch=None,
        reaching_definitions: Optional["ReachingDefinitionsModel"] = None,
        immediate_stmt_removal: bool = False,
        bp_as_gpr: bool = False,
    ):
        super().__init__()

        # Used in the VEX engine
        self._project = project
        self.arch = arch
        self.base_state = None
        self._load_callback = None
        self._propagate_tmps: bool = propagate_tmps
        self._reaching_definitions = reaching_definitions
        self._immediate_stmt_removal = immediate_stmt_removal
        self.bp_as_gpr = bp_as_gpr
        self.stmts_to_remove = set()

        # Used in the AIL engine
        self._stack_pointer_tracker = stack_pointer_tracker

        self._multi_occurrence_registers = None

    def process(self, state, *args, **kwargs):
        self.project = kwargs.pop("project", None)
        self.base_state = kwargs.pop("base_state", None)
        self._load_callback = kwargs.pop("load_callback", None)
        try:
            self._process(state, None, block=kwargs.pop("block", None))
        except SimEngineError as ex:
            if kwargs.pop("fail_fast", False) is True:
                raise ex
            l.error(ex, exc_info=True)

        return self.state



#!/usr/bin/env python3
# pylint:disable=isinstance-second-argument-not-valid-type,missing-class-docstring,no-self-use
import unittest
from unittest import TestCase

import claripy

from angr.storage.memory_mixins.paged_memory.pages.history_tracking_mixin import MAX_HISTORY_DEPTH
from angr.storage.memory_mixins import (
    DataNormalizationMixin,
    SizeNormalizationMixin,
    AddressConcretizationMixin,
    UltraPagesMixin,
    ListPagesMixin,
    PagedMemoryMixin,
    SymbolicMergerMixin,
    ConvenientMappingsMixin,
)
from angr import SimState
from angr.storage.memory_mixins import UltraPage


class UltraPageMemory(
    DataNormalizationMixin,
    SizeNormalizationMixin,
    AddressConcretizationMixin,
    SymbolicMergerMixin,
    ConvenientMappingsMixin,
    UltraPagesMixin,
    PagedMemoryMixin,
):
    pass


class ListPageMemory(
    DataNormalizationMixin,
    SizeNormalizationMixin,
    AddressConcretizationMixin,
    SymbolicMergerMixin,
    ConvenientMappingsMixin,
    ListPagesMixin,
    PagedMemoryMixin,
):
    pass


class TestMemoryMerge(TestCase):
    def test_merge_memory_object_endness(self):
        for memcls in [UltraPageMemory, ListPageMemory]:
            state0 = SimState(arch="AMD64", mode="symbolic", plugins={"memory": memcls()})
            state0.memory.store(0x20000, claripy.BVS("x", 64), endness="Iend_LE")

            state1 = SimState(arch="AMD64", mode="symbolic", plugins={"memory": memcls()})
            state1.memory.store(0x20000, claripy.BVS("y", 64), endness="Iend_LE")

            state, _, _ = state0.merge(state1)
            obj = state.memory.load(0x20000, size=8, endness="Iend_LE")
            assert isinstance(obj, claripy.ast.Base)
            # the original endness should be respected, and obj.op should not be Reverse
            assert obj.op == "If"

    def test_merge_seq(self):
        state1 = SimState(arch="AMD64", mode="symbolic", plugins={"memory": UltraPageMemory()})
        state2 = SimState(arch="AMD64", mode="symbolic", plugins={"memory": UltraPageMemory()})

        state1.regs.rsp = 0x80000000
        state2.regs.rsp = 0x80000000

        state1.memory.store(state1.regs.rsp, 0x11, 1)
        state1.memory.store(state1.regs.rsp + 1, 0x22, 1)
        state2.memory.store(state2.regs.rsp, 0xAA, 1)
        state2.memory.store(state2.regs.rsp + 1, 0xBB, 1)

        state3, _, __ = state1.merge(state2)
        vals = (v for v in state3.solver.eval_upto(state3.memory.load(state3.regs.rsp, 2), 10))
        assert {0x1122, 0xAABB} == set(vals)

    def test_history_tracking(self):
        state = SimState(arch="AMD64", mode="symbolic", plugins={"memory": UltraPageMemory()})

        states = [state]

        for i in range(25):
            state = state.copy()
            states.append(state)  # keep references
            state.memory.store(i, claripy.BVV(i, 8))

        assert len(state.memory._pages) == 1
        page: UltraPage = next(iter(state.memory._pages.values()))

        parents = list(page.parents())
        assert len(parents) == 24

    def test_history_tracking_collapse(self):
        state = SimState(arch="AMD64", mode="symbolic", plugins={"memory": UltraPageMemory()})
        state.memory.store(1000, claripy.BVV(1, 8))

        states = [state]

        for i in range(MAX_HISTORY_DEPTH + 4):
            state = state.copy()
            states.append(state)  # keep references
            state.memory.store(i, claripy.BVV(i, 8))
            assert next(iter(state.memory._pages.values()))._history_depth == (i + 1) % (MAX_HISTORY_DEPTH + 1)

        assert len(state.memory._pages) == 1
        page: UltraPage = next(iter(state.memory._pages.values()))

        parents = list(page.parents())
        assert len(parents) == 3


if __name__ == "__main__":
    unittest.main()



import logging
import angr


l = logging.getLogger(name=__name__)


# pylint:disable=redefined-builtin,arguments-differ
class getrlimit(angr.SimProcedure):
    def run(self, resource, rlim):
        if self.state.solver.eval(resource) == 3:  # RLIMIT_STACK
            l.debug("running getrlimit(RLIMIT_STACK)")
            self.state.memory.store(rlim, 8388608, 8)  # rlim_cur
            self.state.memory.store(
                rlim + 8, self.state.solver.Unconstrained("rlim_max", 8 * 8, key=("api", "rlimit", "stack"))
            )
            return 0
        else:
            l.debug("running getrlimit(other)")
            return self.state.solver.Unconstrained("rlimit", self.arch.sizeof["int"], key=("api", "rlimit", "other"))


class ugetrlimit(getrlimit):
    pass
